import string
import random

import logging
logger = logging.getLogger()


def new_random_string(size=32, chars=string.ascii_uppercase + string.digits):
    """Generate a new random string of size ``size`` containing only the characters provided in ``chars``.

    Parameters
    ----------
    size : int, optional
            The size of the new string. The default value is 32.
    chars : iterable, optional
            A sequence of characters to choose from. The default value is 'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'.

    Returns
    -------
    random_string : str
            The new random string.
    """
    return ('').join((random.choice(chars) for i in range(size)))


def rgb2hex(r, g, b):
    return "#{:02x}{:02x}{:02x}".format(r, g, b)


def hex2rgb(hexcode):
    return tuple(map(ord, hexcode[1:].decode('hex')))


# TODO: Transform hex to rgb and vice versa
class ColorPallete:
    def __init__(self, color_map=None, default_color=(255, 255, 255)):
        self.color_map = color_map or {}
        self.default_color = default_color

    def add_color(self, key, color):
        self.color_map[key] = color

    def get_normalized_color(self, key):
        color = self.get_color(key)

        if isinstance(color, str):
            return color
        elif all([x < 1 for x in color]):
            return color
        return tuple(c / 255 for c in color)

    def get_unnormalized_color(self, key):
        color = self.get_color(key)

        if isinstance(color, str):
            return color
        elif all([x > 1 for x in color]):
            return color
        return tuple(int(c * 255) for c in color)

    def get_color(self, key):
        if key in self.color_map:
            return self.color_map[key]
        elif self.default_color:
            logger.warning("Key '%s' does not exist. Then, the default color will be used: %s." % (key, self.default_color))
            return self.default_color
        else:
            raise KeyError("Key '%s' does not exist and no default color was defined." % key)

    def __contains__(self, key):
        return key in self.color_map


def func_call_to_str(func, *args, **kwargs):
    arg_names = func.__code__.co_varnames[:func.__code__.co_argcount]
    args = args[:len(arg_names)]
    defaults = func.__defaults__ or ()
    args = (args + defaults[len(defaults) - (func.__code__.co_argcount - len(args)):])
    params = zip(arg_names, args)
    args = args[len(arg_names):]
    if args:
        params.append(('args', args))
    if kwargs:
        params.append(('kwargs', kwargs))

    args_as_str = ', '.join('%s=%r' % p for p in params) + ' )'
    func_as_str = "%s(%s)" % (func.__name__, args_as_str)

    return func_as_str


def iter_to_chunks(l, n):
    return [l[i:i + n] for i in range(0, len(l), n)]



from enum import Enum, auto


class IFPType(Enum):
    """An enumeration of Interaction FingerPrints (IFPs) available in LUNA."""

    # auto() creates automatic identifiers for each type as this value is not important.
    # Therefore, always remember to compare IFP types using their names and not their values.
    EIFP = auto()
    HIFP = auto()  # Hybrid version: atomic invariants for atoms and pharmacophore for groups
    FIFP = auto()



import pandas as pd

from luna.util.default_values import LIGAND_EXPO_FILE
from luna.util.exceptions import MoleculeNotFoundError, MoleculeObjectTypeError
from luna.wrappers.base import MolWrapper

from rdkit.Chem.AllChem import AssignBondOrdersFromTemplate

import logging

logger = logging.getLogger()


class Template:
    """Standardize small molecules based on templates."""

    def assign_bond_order(self):
        """Assign bond order to a molecular object. However, this method is not implemented by default.
        Instead, you should use a class that inherits from `Template` and
        implements :meth:`assign_bond_order`. An example is the class `LigandExpoTemplate` that assigns bonds
        order based on ligands SMILES from `LigandExpo <http://ligand-expo.rcsb.org/>`_.
        Therefore, you should define your own logic beyond :meth:`assign_bond_order` that meets your goals."""
        raise NotImplementedError("Subclasses should implement this.")


class LigandExpoTemplate(Template):
    """Standardize small molecules based on templates (SMILES) from `LigandExpo <http://ligand-expo.rcsb.org/>`_.

    Parameters
    ----------
    lig_expo_file : str
        The `LigandExpo <http://ligand-expo.rcsb.org/>`_ file containing the SMILES and ligand ids.
    """

    def __init__(self, lig_expo_file=LIGAND_EXPO_FILE):

        self.lig_expo_file = lig_expo_file
        self._data = None

    @property
    def data(self):
        """:py:class:`pandas.DataFrame` : The `LigandExpo <http://ligand-expo.rcsb.org/>`_ data."""
        if self._data is None:
            self._data = pd.read_csv(self.lig_expo_file, sep="\t+", na_filter=False,
                                     names=["smiles", "ligand_id"], usecols=[0, 1], engine='python').dropna()
        return self._data

    def get_ligand_smiles(self, lig_id):
        """Get SMILES for the ligand ``lig_id``.

        Parameters
        ----------
        lig_id : str
            The ligand identifier (PDB id) in `LigandExpo <http://ligand-expo.rcsb.org/>`_.

        Returns
        ----------
        smiles : str or None
            The ligand SMILES.
        """

        data = self.data[self.data["ligand_id"] == lig_id]["smiles"]
        if data.shape[0] == 0:
            return None
        return data.values[0]

    def assign_bond_order(self, mol_obj, lig_id):
        """Assign bond order to a molecular object based on its `LigandExpo <http://ligand-expo.rcsb.org/>`_ SMILES.

        Parameters
        ----------
        mol_obj : :class:`~luna.wrappers.base.MolWrapper`, :class:`rdkit.Chem.rdchem.Mol`, or :class:`openbabel.pybel.Molecule`
            A molecule to standardise.
        lig_id : str
            The ligand identifier (PDB id) in `LigandExpo <http://ligand-expo.rcsb.org/>`_.

        Returns
        -------
        new_mol : :class:`~luna.wrappers.base.MolWrapper`, :class:`rdkit.Chem.rdchem.Mol`, or :class:`openbabel.pybel.Molecule`
            A standardized molecular object of the same type as ``mol_obj``.
        """

        tmp_mol_obj = MolWrapper(mol_obj)
        if tmp_mol_obj.is_rdkit_obj():
            smiles = self.get_ligand_smiles(lig_id)
            # Raise an exception when the template is not found.
            if smiles is None:
                raise MoleculeNotFoundError("It is not possible to assign the bond orders to the ligand %s because "
                                            "its corresponding template was not found at Ligand Expo." % lig_id)

            template = MolWrapper.from_smiles(smiles, mol_obj_type="rdkit").unwrap()

            # Note that the template molecule should have no explicit hydrogens else the algorithm will fail.
            new_mol = AssignBondOrdersFromTemplate(template, tmp_mol_obj.unwrap())

            if isinstance(mol_obj, MolWrapper):
                return MolWrapper(mol_obj)
            return new_mol
        else:
            logger.exception("Objects of type '%s' are not currently accepted." % mol_obj.__class__)
            raise MoleculeObjectTypeError("Objects of type '%s' are not currently accepted." % mol_obj.__class__)



import logging
logger = logging.getLogger()


class InteractionConfig(dict):
    """Generic class to define parameters for interactions.

    Parameters
    ----------
    config : dict, optional
        A dict containing parameters for calculating interactions.

    Examples
    --------

    >>> config = {"max_ha_dist_hb_inter": 2.5}
    >>> inter_config = InteractionConfig(config)
    >>> print(inter_config["max_ha_dist_hb_inter"])
    2.5
    inter_config["max_ha_dist_hb_inter"] = 3
    >>> print(inter_config["max_ha_dist_hb_inter"])
    3
    """

    def __init__(self, config=None):
        config = config or {}
        super().__init__(config)

    @property
    def params(self):
        """list: The list of parameters."""
        return sorted(self.keys())


class DefaultInteractionConfig(InteractionConfig):
    """Default parameters for calculating interactions in LUNA."""

    def __init__(self):
        config = {}

        # Hydrogen bond
        config["max_da_dist_hb_inter"] = 3.9
        config["max_ha_dist_hb_inter"] = 2.8
        config["min_dha_ang_hb_inter"] = 90
        config["min_har_ang_hb_inter"] = 90
        config["min_dar_ang_hb_inter"] = 90

        # Weak hydrogen bond
        # Ref: Panigrahi, S. K. & Desiraju, G. R. (2007).
        # Ref: Desiraju, G. R. & Steiner, T. (2001).
        config["max_da_dist_whb_inter"] = 4
        config["max_ha_dist_whb_inter"] = 3
        config["min_dha_ang_whb_inter"] = 110
        config["min_har_ang_whb_inter"] = 90
        config["min_dar_ang_whb_inter"] = 90

        # Weak hydrogen bonds: hydrogen bonds involving aromatic rings
        # Ref: Hydrogen bonds with π-acceptors in proteins: frequencies and role in stabilizing local 3D structures [Steiner, 2001]
        # Ref: Strong and Weak Hydrogen Bonds [Panigrahi, 2007]
        config["max_dc_dist_whb_inter"] = 4.5
        config["max_hc_dist_whb_inter"] = 3.5
        config["min_dhc_ang_whb_inter"] = 120
        config["max_disp_ang_whb_inter"] = 40

        # Ionic interactions
        # Ref: Barlow, D. J.; Thornton, J. M. Ion-Pairs in Proteins. Journal of Molecular Biology 1983.
        config["max_dist_repuls_inter"] = 6
        config["max_dist_attract_inter"] = 6

        # Aromatic stacking:
        # Ref: Bhattacharyya et al. 2003. Geometry of Interaction of the Histidine Ring with Other Planar and Basic Residues.
        # Ref: Chakrabarti, P.; Bhattacharyya, R. 2007. Geometry of Nonbonded Interactions Involving Planar Groups in Proteins.
        config["max_cc_dist_pi_pi_inter"] = 6
        # Define if a ring is tilted, same plane or in a T-format
        config["min_dihed_ang_slope_pi_pi_inter"] = 30
        config["max_dihed_ang_slope_pi_pi_inter"] = 60
        # Define if the ring is offset or aligned with another ring center.
        config["min_disp_ang_offset_pi_pi_inter"] = 30
        config["max_disp_ang_offset_pi_pi_inter"] = 60

        # Amide-aromatic stacking
        # [1] A systematic analysis of atomic protein–ligand interactions in the PDB [Freitas, 2007].
        # [2] Efficient Stacking on Protein Amide Fragments [Harder, 2013].
        # [3] The environment of amide groups in protein–ligand complexes: H-bonds and beyond [Cotesta, 2006].
        # [4] Hydrogen bonds with π-acceptors in proteins: frequencies and role in stabilizing local 3D structures [Steiner, 2001]
        # [5] Example of amide-pi: Systematic Investigation of Halogen Bonding in Protein–Ligand Interactions [Hardegger, 2011]
        #
        # Ref: [1], [2], [4], and [5].
        config["max_cc_dist_amide_pi_inter"] = 4.5
        # Ref: [1] and [3].
        config["max_dihed_ang_amide_pi_inter"] = 30
        # Ref: [3]: I use the centroid of an amide, while they use the nitrogen.
        config["max_disp_ang_pi_pi_inter"] = 30

        # Hydrophobic interaction
        config["max_dist_hydrop_inter"] = 4.5
        config["min_surf_size"] = 1
        config["min_inter_atom_in_surf"] = 1

        # Cation-pi interaction
        # Ref: Gallivan, J. P.; Dougherty, D. A. Cation-π Interactions in Structural Biology. Proc Natl Acad Sci U S A 1999, 96 (17), 9459–9464.
        config["max_dist_cation_pi_inter"] = 6

        # Halogen bond.
        # Interaction model: C-X ---- A-R, where C is a carbon, X a halogen, A an acceptor and R is an atom bonded to A.
        # Ref: Halogen bonds in biological molecules [Auffinger, 2004]
        # Ref: The Important Role of Halogen Bond in Substrate Selectivity of Enzymatic Catalysis [Jiang, 2016]
        #
        # Distance X-A when A is a single atom.
        config["max_xa_dist_xbond_inter"] = 4
        # Distance X-A when A is an aromatic ring, so C stands for Centroid.
        config["max_xc_dist_xbond_inter"] = 4.5
        config["min_cxa_ang_xbond_inter"] = 120
        config["min_xar_ang_xbond_inter"] = 80
        config["max_disp_ang_xbond_inter"] = 60

        # Chalcogen bond.
        # Interaction model: R-Y ---- A-N,
        # Where R is a carbon/sulfur, Y a chalcogen (S, Se, Te), A an acceptor, and N a covalently bonded atom to A.
        # Ref: Mining and Structural Characterization of S···X Chalcogen Bonds in Protein Database [Iwaoka, Michio, and Natsuki Babe, 2015].
        # Ref: Chalcogen Bonding ‘2S–2N Squares’ versus Competing Interactions: Exploring the Recognition Properties of Sulfur [Ams et al, 2018].
        # Ref: S···O and S···N Sulfur Bonding Interactions in Protein–Ligand Complexes:
        #       Empirical Considerations and Scoring Function [Koebel, 2016].
        #
        # Distance Y-A when A is a single atom.
        config["max_ya_dist_ybond_inter"] = 4
        # Distance Y-A, when A (acceptor) is an aromatic ring, so C stands for Centroid.
        config["max_yc_dist_ybond_inter"] = 4.5
        # Angles
        config["min_rya_ang_ybond_inter"] = 120
        config["min_yan_ang_ybond_inter"] = 80
        # When A (acceptor) is an aromatic ring.
        config["max_disp_ang_ybond_inter"] = 60

        # Orthogonal multipolar interaction (dipole-dipole)
        # All rules were obtained departing from Paulini R, Müller K, Diederich F. 2005.
        #       Orthogonal multipolar interactions in structural chemistry and biology.
        # Model: A-N ... E-Y, where N and E are the nucleophilic and electrophilic atom, respectively.
        # Distance between the nucleophilic and electrophilic atom.
        config["max_ne_dist_multipolar_inter"] = 4
        # The angle NEY have a min and a max angle.
        config["min_ney_ang_multipolar_inter"] = 70
        config["max_ney_ang_multipolar_inter"] = 110
        config["max_disp_ang_multipolar_inter"] = 40
        # Orthogonal multipolar
        config["min_an_ey_ang_ortho_multipolar_inter"] = 70
        config["max_an_ey_ang_ortho_multipolar_inter"] = 110
        # Parallel and antiparallel multipolar
        config["max_an_ey_ang_para_multipolar_inter"] = 25
        config["min_an_ey_ang_antipara_multipolar_inter"] = 155

        # Ion-multipole interaction (ion-dipole)
        # Model: I ... D-Y, where I is the ion, D the dipole atom of interest (the electrophile or nucleophile) and Y is its counterpart.
        config["max_id_dist_ion_multipole_inter"] = 4.5
        config["min_idy_ang_ion_multipole_inter"] = 60
        config["max_disp_ang_ion_multipole_inter"] = 40

        # Proximal interactions
        config["max_dist_proximal"] = 6
        config["min_dist_proximal"] = 2

        # Van der Waals interactions. From: Arpeggio.
        config["vdw_tolerance"] = 0.1

        # From Chimera: https://www.cgl.ucsf.edu/chimerax/docs/user/commands/clashes.html
        config["vdw_clash_tolerance"] = 0.6

        # From Chimera: https://www.cgl.ucsf.edu/chimerax/docs/user/commands/clashes.html
        # Controls all interactions, except clashes that has its own parameter.
        config["min_bond_separation"] = 3
        # Controls only clashes.
        config["min_bond_separation_for_clash"] = 4

        config["boundary_cutoff"] = 6.2

        super().__init__(config)



import numpy as np
import scipy.optimize
import functools
from scipy.spatial import distance


def atom_coordinates(atoms):
    return np.array([x.get_coord() for x in atoms])


def ob_atom_coordinates(ob_atoms):
    return np.array([(a.GetX(), a.GetY(), a.GetZ()) for a in ob_atoms])


def axis_sum(arr, decimals=3):
    values = [0] * arr.shape[1]
    for i in range(0, arr.shape[1]):
        values[i] = np.around(np.sum(arr[:, i]), decimals)
    return np.array(values)


def point_in_line(p1, p2, d):
    v = p2 - p1
    u = v / np.linalg.norm(v)
    return p1 + d * u


def orthog_point(point, normal, t):
    return point + (normal * t)


def centroid(arr, decimals=3):
    return np.around(axis_sum(arr) / arr.shape[0], decimals)


def euclidean_distance(p1, p2, decimals=3):
    return round(distance.euclidean(p1, p2), decimals)


def angle(p1, p2, decimals=3):
    normal_prod = np.linalg.norm(p1) * np.linalg.norm(p2)

    if normal_prod == 0:
        return np.float32(0)

    dot_prod = np.dot(p1, p2)
    cos_angle = dot_prod / normal_prod
    arcos_angle = np.arccos(np.clip(cos_angle, -1, 1))

    return np.around(np.degrees(arcos_angle), 3)


def norm_vector(p1, p2):
    p1 = p1 / np.linalg.norm(p1)
    p2 = p2 / np.linalg.norm(p2)
    return p2 - p1


def to_quad1(angle, decimals=3):
    if (angle > 90 and angle <= 180):
        return np.around((180 - angle), decimals)
    elif (angle > 180 and angle <= 270):
        return np.around((angle - 180), decimals)
    elif (angle > 270 and angle <= 360):
        return np.around((360 - angle), decimals)
    else:
        return np.around(angle, decimals)


def plane(x, y, params):
    a = params[0]
    b = params[1]
    c = params[2]

    return a * x + b * y + c


def error(params, points):
    result = 0
    for (x, y, z) in points:
        plane_z = plane(x, y, params)
        diff = abs(plane_z - z)
        result += diff**2
    return result


def cross(a, b):
    return [a[1] * b[2] - a[2] * b[1],
            a[2] * b[0] - a[0] * b[2],
            a[0] * b[1] - a[1] * b[0]]


def calc_normal(points, decimals=3):
    fun = functools.partial(error, points=points)
    params0 = [0, 0, 0]

    # It was found a Warning when executing the next line.
    # The Warning is:
    #       optimize.py:994: RuntimeWarning:
    #              divide by zero encountered in double_scalars
    #       rhok = 1.0 / (numpy.dot(yk, sk))
    #
    # However, according to my analysis, Numpy is able to deal
    # with such Warning and the result is generated properly.
    res = scipy.optimize.minimize(fun, params0)

    a = res.x[0]
    b = res.x[1]

    xs, ys, zs = zip(*points)

    return np.around(np.array(cross([1, 0, a], [0, 1, b])), decimals)



# Copyright (C) 2002, Thomas Hamelryck (thamelry@binf.ku.dk)
# This code is part of the Biopython distribution and governed by its
# license.  Please see the LICENSE file that should have been included
# as part of this package.

###################################################################

# Modifications included by Alexandre Fassio (alexandrefassio@dcc.ufmg.br).
# Date: 19/02/2018.

# 1) The module __lt__ was overwritten to allow sorting in Python 3
# 2) Inherit inhouse modifications. Package: MyBio.

# Date: 02/22/2019
# 1) Added property "_is_target" to control chains that will be targets for some processing.
# 2) Added function "is_target() to verify the status of the is_target variable.
# 3) Added function "set_as_target()" to allow the definition if a chain is a target or not.

# Each line or block with modifications contain a MODBY tag.

###################################################################


"""Chain class, used in Structure objects."""

# MODBY: Alexandre Fassio
# Inherit inhouse modifications. Package: MyBio.
from luna.MyBio.PDB.Entity import Entity


class Chain(Entity):
    def __init__(self, id):
        self.level = "C"

        # MODBY: Alexandre Fassio
        # By default: no chain is a target for any calculations.
        self._is_target = False

        Entity.__init__(self, id)

    # Special methods

    # MODBY: Alexandre Fassio.
    # __lt__ method overwritten.
    def __lt__(self, c2):
        return self.id < c2.id

    # Private methods

    def _sort(self, r1, r2):
        """Sort function for residues in a chain (PRIVATE).

        Residues are first sorted according to their hetatm records.
        Protein and nucleic acid residues first, hetatm residues next,
        and waters last. Within each group, the residues are sorted according
        to their resseq's (sequence identifiers). Finally, residues with the
        same resseq's are sorted according to icode.

        Arguments:

        - r1, r2 - Residue objects

        """
        hetflag1, resseq1, icode1 = r1.id
        hetflag2, resseq2, icode2 = r2.id
        if hetflag1 != hetflag2:
            return cmp(hetflag1[0], hetflag2[0])
        elif resseq1 != resseq2:
            return cmp(resseq1, resseq2)
        return cmp(icode1, icode2)

    def _translate_id(self, id):
        """Translate sequence identifer to tuple form (PRIVATE).

        A residue id is normally a tuple (hetero flag, sequence identifier,
        insertion code). Since for most residues the hetero flag and the
        insertion code are blank (i.e. " "), you can just use the sequence
        identifier to index a residue in a chain. The _translate_id method
        translates the sequence identifier to the (" ", sequence identifier,
        " ") tuple.

        Arguments:

        - id - int, residue resseq

        """
        if isinstance(id, int):
            id = (' ', id, ' ')
        return id

    # Special methods

    def __getitem__(self, id):
        """Return the residue with given id.

        The id of a residue is (hetero flag, sequence identifier, insertion code).
        If id is an int, it is translated to (" ", id, " ") by the _translate_id
        method.

        Arguments:

        - id - (string, int, string) or int

        """
        id = self._translate_id(id)
        return Entity.__getitem__(self, id)

    def __contains__(self, id):
        """True if a residue with given id is present in this chain.

        Arguments:

        - id - (string, int, string) or int

        """
        id = self._translate_id(id)
        return Entity.__contains__(self, id)

    def __delitem__(self, id):
        """Delete item.

        Arguments:

        - id - (string, int, string) or int

        """
        id = self._translate_id(id)
        return Entity.__delitem__(self, id)

    def __repr__(self):
        return "<Chain id=%s>" % self.get_id()

    # Public methods

    def get_unpacked_list(self):
        """Return a list of undisordered residues.

        Some Residue objects hide several disordered residues
        (DisorderedResidue objects). This method unpacks them,
        ie. it returns a list of simple Residue objects.
        """
        unpacked_list = []
        for residue in self.get_list():
            if residue.is_disordered() == 2:
                for dresidue in residue.disordered_get_list():
                    unpacked_list.append(dresidue)
            else:
                unpacked_list.append(residue)
        return unpacked_list

    def has_id(self, id):
        """Return 1 if a residue with given id is present.

        The id of a residue is (hetero flag, sequence identifier, insertion code).

        If id is an int, it is translated to (" ", id, " ") by the _translate_id
        method.

        Arguments:

        - id - (string, int, string) or int

        """
        id = self._translate_id(id)
        return Entity.has_id(self, id)

    # MODBY: Alexandre Fassio
    # Check if a chain is a target, i.e., if it will be used for any calculations.
    def is_target(self):
        return self._is_target

    # Public

    def get_residues(self):
        for r in self:
            yield r

    def get_atoms(self):
        for r in self.get_residues():
            for a in r:
                yield a

    # MODBY: Alexandre Fassio
    # Define if all the residues in the chain are a target or not.
    def set_as_target(self, is_target=True):
        self._is_target = is_target

        for r in self.get_residues():
            r.set_as_target(is_target)



from subprocess import Popen, PIPE, TimeoutExpired

from openbabel.pybel import informats, outformats

from luna.util.exceptions import FileNotCreated, InvalidFileFormat
from luna.util.default_values import OPENBABEL
from luna.util.file import is_file_valid, validate_file, get_file_format

import logging
logger = logging.getLogger()


def _prep_opts(opts, prefix=""):
    opt_list = []
    if opts is not None:
        for key in opts:
            if len(key) > 1:
                opt_list.append("--%s%s" % (prefix, key))
            else:
                opt_list.append("-%s%s" % (prefix, key))

            if opts[key] is not None:
                opt_list.append(str(opts[key]))
    return opt_list


def mol_to_svg(infile, output, opts=None):
    """Depict a molecule as SVG using Open Babel.

    Parameters
    ----------
    infile : str
        The pathname of a molecular file.
    output : str
        Save the SVG to this file.
    opts : dict
        A set of depiction options. Check `Open Babel <http://openbabel.org/docs/dev/FileFormats/SVG_depiction.html>`_ \
        to discover which options are available.

    Examples
    --------

    In this example, we will depict the molecule ZINC000007786517 as SVG.
    The following options will be used:
        * C: do not draw terminal C (and attached H) explicitly;
        * d: do not display the molecule name;
        * P: image size in pixels.

    >>> from luna.util.default_values import LUNA_PATH
    >>> from luna.wrappers.obabel import mol_to_svg
    >>> mol_to_svg(infile=f"{LUNA_PATH}/tutorial/inputs/ZINC000007786517.mol",
    ...            output="example.svg",
    ...            opts={"C": None, "d": None, "P": 500})

    """

    logger.debug("Trying to generate a 2D diagram to the file '%s'." % infile)

    try:
        validate_file(infile)

        opt_list = _prep_opts(opts, "x")
        args = ["obabel", infile, "-O", output] + opt_list

        p = Popen(args, stdout=PIPE, stderr=PIPE)
        stdout, stderr = p.communicate()

        svg_created = stderr.decode().strip() == "1 molecule converted"
        if not is_file_valid(output) or not svg_created:
            raise FileNotCreated("SVG diagram for '%s' not created." % infile)
        else:
            logger.debug("SVG diagram for '%s' created." % infile)
    except Exception as e:
        logger.exception(e)
        raise


def convert_molecule(infile, output, infile_format=None,
                     output_format=None, opts=None, openbabel=OPENBABEL):
    """Convert a molecular file to another format using Open Babel.

    Parameters
    ----------

    infile : str
        The pathname of the molecular file to be converted.
    output : str
        Save the converted molecule to this file.
    infile_format : str, optional
        The molecular format of ``infile``.
        If not provided, the format will be defined by the file extension.
    output_format : str, optional
        The molecular format of ``output``.
        If not provided, the format will be defined by the file extension.
    opts : dict
        A set of convertion options. Check `Open Babel <https://openbabel.org/docs/dev/Command-line_tools/babel.html>`_ to discover which options are available.
    openbabel : str, optional
        The Open Babel binary location. If not provided, the default binary ('obabel') will be used.

    Raises
    ------
    InvalidFileFormat
        If the provided molecular formats are not accepted by Open Babel.

    Examples
    --------

    In this example, we will convert the molecule ZINC000007786517 from the format MOL to MOL2
    and add hydrogens to it considering a pH of 7 (option "p").

    >>> from luna.util.default_values import LUNA_PATH
    >>> from luna.wrappers.obabel import convert_molecule
    >>> convert_molecule(infile=f"{LUNA_PATH}/tutorial/inputs/ZINC000007786517.mol",
    ...                  output="example.mol2",
    ...                  opts={"p": 7})
    """
    logger.debug("Trying to convert the file '%s' to '%s'." % (infile, output))

    try:
        # It raises an error if it is not valid.
        validate_file(infile)

        if infile_format is None:
            logger.debug("Input file format not defined. It will assume the format from the file extension.")
            infile_format = get_file_format(infile)

        if infile_format not in informats:
            raise InvalidFileFormat("Infile format '%s' does not exist." % infile_format)

        logger.debug("Input format: %s" % infile_format)

        if output_format is None:
            logger.debug("Output file format not defined. It will assume the format by the file extension.")
            output_format = get_file_format(output, 1)

        if output_format not in outformats:
            raise InvalidFileFormat("Infile format '%s' does not exist." % output_format)

        logger.debug("Output format: %s" % output_format)

        opt_list = _prep_opts(opts)
        args = [openbabel, "-i", infile_format, infile, "-o", output_format, "-O", output] + opt_list

        p = Popen(args, stdout=PIPE, stderr=PIPE)
        try:
            stdout, stderr = p.communicate(timeout=30)
        except TimeoutExpired:
            p.kill()
            raise

        output_lines = stderr.decode().strip().split("\n")
        is_mol_converted = output_lines[-1] != "0 molecule converted"
        if not is_file_valid(output) or not is_mol_converted:
            raise FileNotCreated("File '%s' not converted to '%s'." % (infile, output))
        else:
            logger.debug("File '%s' converted to '%s'." % (infile, output))
    except Exception:
        raise



"""
Functions to convert strings.
"""

import re


def camelcase(string):
    """ Convert string into camel case.

    Parameters
    ----------
        string : str
            String to convert.

    Returns
    -------
        string : str
            Camel case string.
    """

    string = re.sub(r"\w[\s\W]+\w", '', str(string))
    if not string:
        return string
    return lowercase(string[0]) + re.sub(r"[\-_\.\s]([a-z])", lambda matched: uppercase(matched.group(1)), string[1:])


def capitalcase(string):
    """Convert string into capital case.
    First letters will be uppercase.

    Parameters
    ----------
        string : str
            String to convert.

    Returns
    -------
        string : str
            Capital case string.

    Examples
    --------
    >>> import luna.util.stringcase as sc
    >>> new_str = sc.camelcase("This is an interesting example.")
    >>> print(new_str)
    'this is an interesting example.'

    """
    string = str(string)
    if not string:
        return string
    return uppercase(string[0]) + string[1:]


def constcase(string):
    """Convert string into upper snake case.
    Join punctuation with underscore and convert letters into uppercase.

    Parameters
    ----------
        string : str
            String to convert.

    Returns
    -------
        string : str
            Const cased string.
    """

    return uppercase(snakecase(string))


def lowercase(string):
    """Convert string into lower case.

    Parameters
    ----------
        string : str
            String to convert.

    Returns
    -------
        string : str
            Lowercase case string.

    Examples
    --------
    >>> import luna.util.stringcase as sc
    >>> new_str = sc.camelcase("This is an Interesting EXAMPLE.")
    >>> print(new_str)
    'this is an interesting example.'
    """

    return str(string).lower()


def pascalcase(string):
    """Convert string into pascal case.

    Parameters
    ----------
        string : str
            String to convert.

    Returns
    -------
        string : str
            Pascal case string.
    """

    return capitalcase(camelcase(string))


def pathcase(string):
    """Convert string into path case.
    Join punctuation with slash.

    Parameters
    ----------
        string : str
            String to convert.

    Returns
    -------
        string : str
            Path cased string.
    """
    string = snakecase(string)
    if not string:
        return string
    return re.sub(r"_", "/", string)


def backslashcase(string):
    """Convert string into spinal case.
    Join punctuation with backslash.

    Parameters
    ----------
        string : str
            String to convert.

    Returns
    -------
        string : str
            Spinal cased string.
    """
    str1 = re.sub(r"_", r"\\", snakecase(string))

    return str1
    # return re.sub(r"\\n", "", str1))  # TODO: make regex fot \t ...


def sentencecase(string):
    """Convert string into sentence case.
    First letter capped and each punctuations are joined with space.

    Parameters
    ----------
        string : str
            String to convert.

    Returns
    -------
        string : str
            Sentence cased string.
    """
    joiner = ' '
    string = re.sub(r"[\-_\.\s]", joiner, str(string))
    if not string:
        return string
    return capitalcase(trimcase(
        re.sub(r"[A-Z]", lambda matched: joiner +
               lowercase(matched.group(0)), string)
    ))


def snakecase(string):
    """Convert string into snake case.
    Join punctuation with underscore.

    Parameters
    ----------
        string : str
            String to convert.

    Returns
    -------
        string : str
            Snake cased string.
    """

    string = re.sub(r"[\-\.\s]", '_', str(string))
    if not string:
        return string
    return lowercase(string[0]) + re.sub(r"[A-Z]", lambda matched: '_' + lowercase(matched.group(0)), string[1:])


def spinalcase(string):
    """Convert string into spinal case.
    Join punctuation with hyphen.

    Parameters
    ----------
        string : str
            String to convert.

    Returns
    -------
        string : str
            Spinal cased string.
    """

    return re.sub(r"_", "-", snakecase(string))


def dotcase(string):
    """Convert string into dot case.
    Join punctuation with dot.

    Parameters
    ----------
        string : str
            String to convert.

    Returns
    -------
        string : str
            Dot cased string.
    """

    return re.sub(r"_", ".", snakecase(string))


def titlecase(string):
    """Convert string into sentence case.
    First letter capped while each punctuations is capitalsed
    and joined with space.

    Parameters
    ----------
        string : str
            String to convert.

    Returns
    -------
        string : str
            Title cased string.
    """

    return ' '.join(
        [capitalcase(word) for word in snakecase(string).split("_")]
    )


def trimcase(string):
    """Convert string into trimmed string.

    Parameters
    ----------
        string : str
            String to convert.

    Returns
    -------
        string : str
            Trimmed case string.

    Examples
    --------
    >>> import luna.util.stringcase as sc
    >>> new_str = sc.camelcase("      This is an Interesting EXAMPLE.     ")
    >>> print(new_str)
    'This is an Interesting EXAMPLE.'
    """
    return str(string).strip()


def uppercase(string):
    """Convert string into upper case.

    Parameters
    ----------
        string : str
            String to convert.

    Returns
    -------
        string : str
            Uppercase case string.

    Examples
    --------
    >>> import luna.util.stringcase as sc
    >>> new_str = sc.camelcase("This is an Interesting EXAMPLE.")
    >>> print(new_str)
    'THIS IS AN INTERESTING EXAMPLE.'
    """
    return str(string).upper()


def alphanumcase(string):
    """Cuts all non-alphanumeric symbols,
    i.e. cuts all expect except 0-9, a-z and A-Z.

    Parameters
    ----------
        string : str
            String to convert.

    Returns
    -------
        string : str
            String with cutted non-alphanumeric symbols.
    """
    # return filter(str.isalnum, str(string))
    return re.sub("\W+", "", string)



from enum import Enum, auto

from luna.wrappers.base import BondType


import logging

logger = logging.getLogger()


METALS = ["Li", "Na", "K", "Rb", "Cs", "Fr", "Be", "Mg", "Ca", "Sr", "Ba", "Ra", "Sc", "Ti", "V", "Cr", "Mn", "Fe",
          "Co", "Ni", "Cu", "Zn", "Al", "Ga", "Y", "Zr", "Nb", "Mo", "Tc", "Ru", "Rh", "Pd", "Ag", "Cd", "In", "Sn", "Hf",
          "Ta", "W", "Re", "Os", "Ir", "Pt", "Au", "Hg", "Tl", "Pb", "Bi"]
METAL_ATOM = "[%s]" % ",".join(METALS)


class ResidueStandard(Enum):
    """An enumeration of protonation states for standard residues."""

    # TODO: add other residues to the standard

    # HID as in Amber: histidine with hydrogen on the delta nitrogen.
    HID = auto()
    # HIE as in Amber: histidine with hydrogen on the epsilon nitrogen.
    HIE = auto()
    # HIP as in Amber: histidine with hydrogens on both nitrogens; this is positively charged.
    HIP = auto()

    # CYS as in Amber: protonated cysteine
    CYS = auto()
    # CYM as in Amber: deprotonated cysteine.
    CYM = auto()


# TODO: create a new MetalStandardiser or LigandStandardiser for common ligands.

class ResiduesStandardiser:
    """Standardize residues.

    Parameters
    ----------
    break_metal_bonds : bool
        If True, break covalent bonds with metals and correct the topology of the involved atoms.
    his_type : {:class:`~ResidueStandard.HID`, :class:`~ResidueStandard.HIE`, :class:`~ResidueStandard.HIP`}
        Define which histidine protonation state to use. Currently, this option is still not been used.

    """

    # TODO: create filters for specific patterns like HIS tautomers or CYS:S- vs CYS:SH

    # TODO: some residues still have no WARNING for unexpected atoms bound to the atoms N, O, S.

    # TODO: I need to verify for HIS tautomers and metals around it, which will influenciate on the Hydrogen placement.

    # TODO: Metals will influentiate where the Hydrogen should be placed in TYR.

    def __init__(self, break_metal_bonds=False, his_type=ResidueStandard.HIE):

        self.break_metal_bonds = break_metal_bonds

        # TODO: not being used
        self.his_type = his_type

    def standardise(self, atom_pairs):
        """Standardize residues.

        Parameters
        ----------
        atom_pairs : iterable of tuple of (:class:`~luna.wrappers.base.MolWrapper`, :class:`~luna.MyBio.PDB.Atom.Atom`)
            The atoms to standardise.
        """
        pdb_map = {atm_obj.get_idx(): pdb_atm for atm_obj, pdb_atm in atom_pairs}

        self.found_metals = {}
        self.removed_atoms = []

        for atm_obj, pdb_atm in atom_pairs:

            # Atom: N
            # Sanity check for all N nitrogens with invalid bond types.
            #
            # E.g.: 3QQL:A:GLY:11.
            if pdb_atm.name == "N":

                # Non-N-terminal N where the carbonyl O is involved in a covalent bond with an atom not comprised
                # by the metal rules. So, it's better not to update anything here.
                if atm_obj.matches_smarts(f"N-,=[C;X4,X3]([#6])[OX2;$(O([C;X4,X3])[!#1]);!$(O{METAL_ATOM})]"):
                    logger.debug("While checking for inconsistencies in the atom N of the residue %s, it was found an unexpected "
                                 "atom covalently bound to the neighboring atom O. So, N will not be amended." % pdb_atm.parent)
                    continue
                # Non-N-terminal N where the carbonyl O is involved in a covalent bond with a metal.
                # However, 'break_metal_bonds' was set to False and there isn't anything to be done so.
                elif atm_obj.matches_smarts(f"N-,=[C;X4,X3]([#6])[OX2]{METAL_ATOM}") and self.break_metal_bonds is False:
                    logger.debug("While checking for inconsistencies in the atom N of the residue %s, it was found a metal "
                                 "covalently bound to its atom O. However, nothing will be done because 'break_metal_bonds' "
                                 "was set to False." % pdb_atm.parent)
                    continue

                # Any non-N-terminal N not from PRO.
                if (pdb_atm.parent.resname != "PRO" and atm_obj.get_neighbors_number(True) == 2
                    and (atm_obj.get_valence() != 3 or atm_obj.get_charge() != 0
                         or atm_obj.has_only_bond_type(BondType.SINGLE) is False
                         or atm_obj.get_h_count() != 1 or atm_obj.is_in_ring() or atm_obj.is_aromatic())):

                    self._fix_atom(atm_obj, bond_types=[BondType.SINGLE], charge=0, implicit_h_count=1)

                # Any N-terminal N not from PRO.
                elif (pdb_atm.parent.resname != "PRO" and atm_obj.get_neighbors_number(True) == 1
                        and (atm_obj.get_valence() != 4 or atm_obj.get_charge() != 1
                             or atm_obj.has_only_bond_type(BondType.SINGLE) is False
                             or atm_obj.get_h_count() != 3 or atm_obj.is_in_ring() or atm_obj.is_aromatic())):

                    self._fix_atom(atm_obj, bond_types=[BondType.SINGLE], charge=1, implicit_h_count=3)

                # Non-N-terminal N from PRO.
                elif (pdb_atm.parent.resname == "PRO" and atm_obj.get_neighbors_number(True) == 3
                        and (atm_obj.get_valence() != 3 or atm_obj.get_charge() != 0
                             or atm_obj.has_only_bond_type(BondType.SINGLE) is False
                             or atm_obj.get_h_count() != 0 or atm_obj.is_in_ring() is False or atm_obj.is_aromatic())):

                    self._fix_atom(atm_obj, bond_types=[BondType.SINGLE], charge=0, implicit_h_count=0, in_ring=True)

                # N-terminal N from PRO.
                elif (pdb_atm.parent.resname == "PRO" and atm_obj.get_neighbors_number(True) == 2
                        and (atm_obj.get_valence() != 4 or atm_obj.get_charge() != 1
                             or atm_obj.has_only_bond_type(BondType.SINGLE) is False
                             or atm_obj.get_h_count() != 2 or atm_obj.is_in_ring() is False or atm_obj.is_aromatic())):

                    self._fix_atom(atm_obj, bond_types=[BondType.SINGLE], charge=1, implicit_h_count=2, in_ring=True)

            # Atom: C.
            # Sanity check for all C carbons with invalid bond types.
            elif pdb_atm.name == "C":

                fix_atom = False

                # If it is bonded to 3 heavy atoms.
                if atm_obj.get_neighbors_number(True) == 3:
                    # Main chain carbonyl O involved in a metallic coordination.
                    #
                    #   It identifies only the form generated by Open Babel, where the double bond between O and C becomes
                    #       a single bond, and it adds an additional single bond with a Metal.
                    #
                    #   Note that if the list of atoms is sorted in a way that O comes first than C in this loop, then the
                    #       double bond would have already been amended and the bond with the metal would have already been removed.
                    #
                    #   E.g.: 6JWU:A:GLU:42.
                    #         1DKA:A:PRO:99.
                    #
                    if atm_obj.matches_smarts(f"[C;X4,X3]([$([OX2]{METAL_ATOM})])([#6])(-,=[N,O])"):
                        # If it is necessary to break covalent bonds with metals.
                        if self.break_metal_bonds:
                            fix_atom = True
                        else:
                            logger.debug("While checking for inconsistencies in the atom C of the residue %s, it was found a metal "
                                         "covalently bound to its atom O. However, nothing will be done because 'break_metal_bonds' "
                                         "was set to False." % pdb_atm.parent)

                    # If the neighboring oxygen is bound to something else not comprised in the previous rule, it is better not to
                    # update anything. Otherwise, fix the C.
                    elif (atm_obj.matches_smarts(f"[C;X4,X3]([OX2;$(O([C;X4,X3])[!#1]);!$(O{METAL_ATOM})])([#6])(-,=[N,O])") is False
                            and atm_obj.matches_smarts("[CX3](=[OX1])([#6])[NX3]") is False
                            and atm_obj.matches_smarts("[CX3](=[OX1])([#6])[O;H1,H0&-1]") is False):
                        fix_atom = True

                    # Alert for unexpected atoms bound to the oxygen O.
                    elif atm_obj.matches_smarts(f"[C;X4,X3]([OX2;$(O([C;X4,X3])[!#1]);!$(O{METAL_ATOM})])([#6])(-,=[N,O])"):
                        logger.debug("While checking for inconsistencies in the atom C of the residue %s, it was found an unexpected "
                                     "atom covalently bound to its atom O. So, C will not be amended." % pdb_atm.parent)

                    if fix_atom:
                        bond_types = []
                        for bond_obj in atm_obj.get_bonds():
                            partner_obj = bond_obj.get_partner_atom(atm_obj)
                            if partner_obj.get_idx() in pdb_map and pdb_map[partner_obj.get_idx()].name == "O":
                                bond_types.append((bond_obj, BondType.DOUBLE))
                            else:
                                bond_types.append((bond_obj, BondType.SINGLE))

                        self._fix_atom(atm_obj, bond_types=bond_types, charge=0, implicit_h_count=0)

                # If it is bonded only to 2 heavy atoms, then it is a C-terminal with a missing OXT or there are missing residues.
                elif atm_obj.get_neighbors_number(True) == 2:

                    logger.debug("The residue %s seems not to have any successor residue. "
                                 "It may be the last in the chain sequence or there are missing residues." % pdb_atm.parent)

                    # Main chain carbonyl O involved in a metallic coordination.
                    #
                    #   It identifies only the form generated by Open Babel, where the double bond between O and C becomes
                    #       a single bond, and it adds an additional single bond with a Metal.
                    #
                    #   Note that if the list of atoms is sorted in a way that O comes first than C in this loop, then the
                    #       double bond would have already been amended and the bond with the metal would have already been removed.
                    #
                    #   E.g.: 6JWU:A:GLU:42.
                    #         1DKA:A:THR:98.
                    #
                    if atm_obj.matches_smarts(f"[CX4]([$([OX2]{METAL_ATOM})])([#6])"):
                        # If it is necessary to break covalent bonds with metals.
                        if self.break_metal_bonds:
                            fix_atom = True
                        else:
                            logger.debug("While checking for inconsistencies in the atom C of the residue %s, it was found a metal "
                                         "covalently bound to its atom O. However, nothing will be done because 'break_metal_bonds' "
                                         "was set to False." % pdb_atm.parent)

                    # If the neighboring oxygen is bound to something else not comprised in the previous rule, it is better not to
                    # update anything. Otherwise, fix the C.
                    elif (atm_obj.matches_smarts(f"[CX4]([OX2;$(O([CX4])[!#1]);!$(O{METAL_ATOM})])([#6])") is False and
                            atm_obj.matches_smarts("[CX3](=[OX1])([#6])") is False):
                        fix_atom = True

                    # Alert for unexpected atoms bound to the oxygen O.
                    elif atm_obj.matches_smarts(f"[CX4]([OX2;$(O([CX4])[!#1]);!$(O{METAL_ATOM})])([#6])"):
                        logger.debug("While checking for inconsistencies in the atom C of the residue %s, it was found an unexpected "
                                     "atom covalently bound to its atom O. So, C will not be amended." % pdb_atm.parent)

                    if fix_atom:
                        bond_types = []
                        for bond_obj in atm_obj.get_bonds():
                            partner_obj = bond_obj.get_partner_atom(atm_obj)
                            if partner_obj.get_idx() in pdb_map and pdb_map[partner_obj.get_idx()].name == "O":
                                bond_types.append((bond_obj, BondType.DOUBLE))
                            else:
                                bond_types.append((bond_obj, BondType.SINGLE))

                        self._fix_atom(atm_obj, bond_types=bond_types, charge=0, implicit_h_count=1)

            # Atom: CA
            # Sanity check for all CA with invalid bond types.
            # E.g.: 3QQL:A:GLY:11.
            elif pdb_atm.name == "CA":

                # Any CA not from GLY/PRO.
                if (pdb_atm.parent.resname not in ["GLY", "PRO"] and atm_obj.get_neighbors_number(True) == 3 and
                        (atm_obj.get_valence() != 4 or atm_obj.get_charge() != 0 or atm_obj.has_only_bond_type(BondType.SINGLE) is False or
                            atm_obj.get_h_count() != 1 or atm_obj.is_in_ring() or atm_obj.is_aromatic())):

                    self._fix_atom(atm_obj, bond_types=[BondType.SINGLE], charge=0, implicit_h_count=1)

                elif (pdb_atm.parent.resname == "PRO" and atm_obj.get_neighbors_number(True) == 3 and
                        (atm_obj.get_valence() != 4 or atm_obj.get_charge() != 0 or atm_obj.has_only_bond_type(BondType.SINGLE) is False or
                            atm_obj.get_h_count() != 1 or atm_obj.is_in_ring() is False or atm_obj.is_aromatic())):

                    self._fix_atom(atm_obj, bond_types=[BondType.SINGLE], charge=0, implicit_h_count=1, in_ring=True)

                # CA from GLY.
                elif (pdb_atm.parent.resname == "GLY" and atm_obj.get_neighbors_number(True) == 2 and
                        (atm_obj.get_valence() != 4 or atm_obj.get_charge() != 0 or atm_obj.has_only_bond_type(BondType.SINGLE) is False or
                            atm_obj.get_h_count() != 2 or atm_obj.is_in_ring() or atm_obj.is_aromatic())):

                    self._fix_atom(atm_obj, bond_types=[BondType.SINGLE], charge=0, implicit_h_count=2)

            # Atom: CB
            # Sanity check for all CB carbons with invalid bond types.
            # E.g.: 1THA:B:GLN:63.
            elif pdb_atm.name == "CB":

                # Any CB not from ALA, ILE, PRO, THR, VAL.
                if (pdb_atm.parent.resname not in ["ALA", "ILE", "PRO", "THR", "VAL"] and atm_obj.get_neighbors_number(True) == 2 and
                        (atm_obj.get_valence() != 4 or atm_obj.get_charge() != 0 or atm_obj.has_only_bond_type(BondType.SINGLE) is False or
                            atm_obj.get_h_count() != 2 or atm_obj.is_in_ring() or atm_obj.is_aromatic())):

                    self._fix_atom(atm_obj, bond_types=[BondType.SINGLE], charge=0, implicit_h_count=2)

                # CB from ILE, THR, VAL.
                elif (pdb_atm.parent.resname in ["ILE", "THR", "VAL"] and atm_obj.get_neighbors_number(True) == 3 and
                        (atm_obj.get_valence() != 4 or atm_obj.get_charge() != 0 or atm_obj.has_only_bond_type(BondType.SINGLE) is False or
                            atm_obj.get_h_count() != 1 or atm_obj.is_in_ring() or atm_obj.is_aromatic())):

                    self._fix_atom(atm_obj, bond_types=[BondType.SINGLE], charge=0, implicit_h_count=1)

                # CB from ALA.
                elif (pdb_atm.parent.resname == "ALA" and atm_obj.get_neighbors_number(True) == 1 and
                        (atm_obj.get_valence() != 4 or atm_obj.get_charge() != 0 or atm_obj.has_only_bond_type(BondType.SINGLE) is False or
                            atm_obj.get_h_count() != 3 or atm_obj.is_in_ring() or atm_obj.is_aromatic())):

                    self._fix_atom(atm_obj, bond_types=[BondType.SINGLE], charge=0, implicit_h_count=3)

                # CB from PRO.
                elif (pdb_atm.parent.resname == "PRO" and atm_obj.get_neighbors_number(True) == 2 and
                        (atm_obj.get_valence() != 4 or atm_obj.get_charge() != 0 or atm_obj.has_only_bond_type(BondType.SINGLE) is False or
                            atm_obj.get_h_count() != 2 or atm_obj.is_in_ring() is False or atm_obj.is_aromatic())):

                    self._fix_atom(atm_obj, bond_types=[BondType.SINGLE], charge=0, implicit_h_count=2, in_ring=True)

            # Atom: CG
            # Sanity check for all CG carbons with invalid bond types.
            elif pdb_atm.name == "CG":

                # E.g.: 1THA:B:GLN:63, 6COD:A:LYS:121.
                if (pdb_atm.parent.resname in ["ARG", "GLN", "GLU", "LYS", "MET"] and atm_obj.get_neighbors_number(True) == 2 and
                        (atm_obj.get_valence() != 4 or atm_obj.get_charge() != 0 or atm_obj.has_only_bond_type(BondType.SINGLE) is False or
                            atm_obj.get_h_count() != 2 or atm_obj.is_in_ring() or atm_obj.is_aromatic())):

                    self._fix_atom(atm_obj, bond_types=[BondType.SINGLE], charge=0, implicit_h_count=2)

                elif (pdb_atm.parent.resname == "PRO" and atm_obj.get_neighbors_number(True) == 2 and
                        (atm_obj.get_valence() != 4 or atm_obj.get_charge() != 0 or atm_obj.has_only_bond_type(BondType.SINGLE) is False or
                            atm_obj.get_h_count() != 2 or atm_obj.is_in_ring() is False or atm_obj.is_aromatic())):

                    self._fix_atom(atm_obj, bond_types=[BondType.SINGLE], charge=0, implicit_h_count=2, in_ring=True)

                elif (pdb_atm.parent.resname in ["ASN", "ASP"] and atm_obj.get_neighbors_number(True) == 3 and
                        (atm_obj.get_valence() != 4 or atm_obj.get_charge() != 0 or atm_obj.get_degree() != 3 or
                            atm_obj.get_h_count() != 0 or atm_obj.is_in_ring() or atm_obj.is_aromatic())):

                    fix_atom = False
                    # ASN/ASP with metallic coordination perceived as covalent bond.
                    #
                    #   It identifies only the form generated by Open Babel, where the double bond between OD1 and CG becomes
                    #       a single bond, and it adds an additional single bond with a Metal.
                    #
                    #   Note that if the list of atoms is sorted in a way that OD1 comes first than CG in this loop, then the
                    #       double bond would have already been amended and the bond with the metal would have already been removed.
                    #
                    #   E.g.: 6JWU:B:ASP:210, 4FVR:A:ASN:678
                    #
                    if (atm_obj.matches_smarts(f"[CX4](N)([#6])[OX2]{METAL_ATOM}") or
                            atm_obj.matches_smarts(f"[C;X4,X3](-,=[O])([#6])[OX2]({METAL_ATOM})")):
                        # If it is necessary to break covalent bonds with metals.
                        if self.break_metal_bonds:
                            fix_atom = True
                        else:
                            logger.debug("While checking for inconsistencies in the atom CG of the residue %s, it was found a metal "
                                         "covalently bound to its atom OD1. However, nothing will be done because 'break_metal_bonds' "
                                         "was set to False." % pdb_atm.parent)

                    # If the neighboring oxygen is bound to something else not comprised in the previous rule, it is better not to
                    # update anything. Otherwise, fix the CG.
                    elif (atm_obj.matches_smarts("[CX4](N)([#6])[OX2][!#1]") is False and
                            atm_obj.matches_smarts("[C;X4,X3](-,=[O])([#6])[OX2]([!#1])") is False):
                        fix_atom = True

                    # Alert for unexpected atoms bound to the neighboring nitrogen/oxygen.
                    elif atm_obj.matches_smarts("[C;X4,X3](-,=[N,O])([#6])[OX2][!#1]"):
                        logger.debug("While checking for inconsistencies in the atom CG of the residue %s, it was found an unexpected "
                                     "atom covalently bound to a neighboring nitrogen/oxygen. So, CG will not be amended."
                                     % pdb_atm.parent)

                    if fix_atom:
                        bond_types = []
                        for bond_obj in atm_obj.get_bonds():
                            partner_obj = bond_obj.get_partner_atom(atm_obj)

                            if partner_obj.get_idx() in pdb_map:
                                # Redundant rules: OD1 and OD2 section already fixes the bonds with CG.
                                if partner_obj.get_idx() in pdb_map and pdb_map[partner_obj.get_idx()].name == "OD1":
                                    # If the OD1 atom is bound to the below metals, update the CG-OD1 bond to CG=OD1.
                                    if partner_obj.matches_smarts(f"[OX2]({METAL_ATOM})"):
                                        bond_types.append((bond_obj, BondType.DOUBLE))
                                    else:
                                        bond_types.append((bond_obj, BondType.SINGLE))

                                # Redundant rules: OD1 and OD2 section already fixes the bonds with CG.
                                elif partner_obj.get_idx() in pdb_map and pdb_map[partner_obj.get_idx()].name == "OD2":
                                    # If the second oxygen (OD1) is bound to some atom not comprised by the expected metals, but OD2 is,
                                    # then update the CG-OD2 bond to CG=OD2.
                                    if partner_obj.matches_smarts(f"[OX2]({METAL_ATOM})[CX4]([#6])"
                                                                  f"[OX2;$(O([CX4])[!#1]);!$(O{METAL_ATOM})]"):
                                        bond_types.append((bond_obj, BondType.DOUBLE))
                                    else:
                                        bond_types.append((bond_obj, BondType.SINGLE))

                                # Any other atom bound to CG
                                else:
                                    bond_types.append((bond_obj, BondType.SINGLE))

                        self._fix_atom(atm_obj, bond_types=bond_types, charge=0, implicit_h_count=0)

                elif (pdb_atm.parent.resname == "LEU" and atm_obj.get_neighbors_number(True) == 3 and
                        (atm_obj.get_valence() != 4 or atm_obj.get_charge() != 0 or atm_obj.has_only_bond_type(BondType.SINGLE) is False or
                            atm_obj.get_h_count() != 1 or atm_obj.is_in_ring() or atm_obj.is_aromatic())):

                    self._fix_atom(atm_obj, bond_types=[BondType.SINGLE], charge=0, implicit_h_count=1)

                elif (pdb_atm.parent.resname in ["HIS", "PHE", "TRP", "TYR"] and atm_obj.get_neighbors_number(True) == 3 and
                        (atm_obj.get_valence() != 4 or atm_obj.get_charge() != 0 or atm_obj.get_degree() != 3 or
                            atm_obj.get_h_count() != 0 or atm_obj.is_aromatic() is False)):

                    bond_types = []
                    for bond_obj in atm_obj.get_bonds():
                        partner_obj = bond_obj.get_partner_atom(atm_obj)
                        if partner_obj.get_idx() in pdb_map:
                            # Single bonds with CB (all)
                            if pdb_map[partner_obj.get_idx()].name == "CB":
                                bond_types.append((bond_obj, BondType.SINGLE))
                            # Single bond with ND1 (only in HIS)
                            elif pdb_map[partner_obj.get_idx()].name == "ND1":
                                bond_types.append((bond_obj, BondType.SINGLE, True))
                            # Double bonds with CD1 (all, except HIS)
                            elif pdb_map[partner_obj.get_idx()].name == "CD1":
                                bond_types.append((bond_obj, BondType.DOUBLE, True))
                            # Bonds with CD2
                            elif pdb_map[partner_obj.get_idx()].name == "CD2":
                                # Double bond as in HIS.
                                if pdb_atm.parent.resname == "HIS":
                                    bond_types.append((bond_obj, BondType.DOUBLE, True))
                                # Single bonds (all, except HIS)
                                else:
                                    bond_types.append((bond_obj, BondType.SINGLE, True))

                    self._fix_atom(atm_obj, bond_types=bond_types, charge=0, implicit_h_count=0, is_aromatic=True)

            # Atom: CG1
            # Sanity check for all CG1 carbons with invalid bond types.
            elif pdb_atm.name == "CG1":
                # CG1 from ILE.
                if (pdb_atm.parent.resname == "ILE" and atm_obj.get_neighbors_number(True) == 2 and
                        (atm_obj.get_valence() != 4 or atm_obj.get_charge() != 0 or atm_obj.has_only_bond_type(BondType.SINGLE) is False or
                            atm_obj.get_h_count() != 2 or atm_obj.is_in_ring() or atm_obj.is_aromatic())):

                    self._fix_atom(atm_obj, bond_types=[BondType.SINGLE], charge=0, implicit_h_count=2)

                # CG1 from VAL.
                elif (pdb_atm.parent.resname == "VAL" and atm_obj.get_neighbors_number(True) == 1 and
                        (atm_obj.get_valence() != 4 or atm_obj.get_charge() != 0 or atm_obj.has_only_bond_type(BondType.SINGLE) is False or
                            atm_obj.get_h_count() != 3 or atm_obj.is_in_ring() or atm_obj.is_aromatic())):

                    self._fix_atom(atm_obj, bond_types=[BondType.SINGLE], charge=0, implicit_h_count=3)

            # Atom: CG2
            # Sanity check for all CG2 carbons with invalid bond types.
            elif (pdb_atm.name == "CG2" and atm_obj.get_neighbors_number(True) == 1 and
                    (atm_obj.get_valence() != 4 or atm_obj.get_charge() != 0 or atm_obj.has_only_bond_type(BondType.SINGLE) is False or
                        atm_obj.get_h_count() != 3 or atm_obj.is_in_ring() or atm_obj.is_aromatic())):

                self._fix_atom(atm_obj, bond_types=[BondType.SINGLE], charge=0, implicit_h_count=3)

            # Atom: CD
            # Sanity check for all CD carbons with invalid bond types.
            elif pdb_atm.name == "CD":

                if (pdb_atm.parent.resname in ["ARG", "LYS"] and atm_obj.get_neighbors_number(True) == 2 and
                        (atm_obj.get_valence() != 4 or atm_obj.get_charge() != 0 or atm_obj.has_only_bond_type(BondType.SINGLE) is False or
                            atm_obj.get_h_count() != 2 or atm_obj.is_in_ring() or atm_obj.is_aromatic())):

                    self._fix_atom(atm_obj, bond_types=[BondType.SINGLE], charge=0, implicit_h_count=2)

                # E.g.: 4FVR:A:PRO:792
                elif (pdb_atm.parent.resname == "PRO" and atm_obj.get_neighbors_number(True) == 2 and
                        (atm_obj.get_valence() != 4 or atm_obj.get_charge() != 0 or atm_obj.has_only_bond_type(BondType.SINGLE) is False or
                            atm_obj.get_h_count() != 2 or atm_obj.is_in_ring() is False or atm_obj.is_aromatic())):

                    self._fix_atom(atm_obj, bond_types=[BondType.SINGLE], charge=0, implicit_h_count=2, in_ring=True)

                # E.g.: 1THA:B:GLN:63
                # E.g.: 6JWU:A:GLU:42 containing metallic cordination.
                elif (pdb_atm.parent.resname in ["GLN", "GLU"] and atm_obj.get_neighbors_number(True) == 3 and
                        (atm_obj.get_valence() != 4 or atm_obj.get_charge() != 0 or atm_obj.get_degree() != 3 or
                            atm_obj.get_h_count() != 0 or atm_obj.is_in_ring() or atm_obj.is_aromatic())):

                    fix_atom = False
                    # GLN/GLU with metallic coordination perceived as covalent bond.
                    #
                    #   It identifies only the form generated by Open Babel, where the double bond between OE1 and CD becomes
                    #       a single bond, and it adds an additional single bond with a Metal.
                    #
                    #   Note that if the list of atoms is sorted in a way that OE1 comes first than CD in this loop, then the
                    #       double bond would have already been amended and the bond with the metal would have already been removed.
                    #
                    if (atm_obj.matches_smarts(f"[CX4](N)([#6])[OX2]{METAL_ATOM}") or
                            atm_obj.matches_smarts(f"[C;X4,X3](-,=[O])([#6])[OX2]({METAL_ATOM})")):
                        # If it is necessary to break covalent bonds with metals.
                        if self.break_metal_bonds:
                            fix_atom = True
                        else:
                            logger.debug("While checking for inconsistencies in the atom CD of the residue %s, it was found a metal "
                                         "covalently bound to its atom OE1. However, nothing will be done because 'break_metal_bonds' "
                                         "was set to False." % pdb_atm.parent)

                    # If the neighboring oxygen is bound to something else not comprised in the previous rule, it is better not to
                    # update anything. Otherwise, fix the CD.
                    elif (atm_obj.matches_smarts("[CX4](N)([#6])[OX2][!#1]") is False and
                            atm_obj.matches_smarts("[C;X4,X3](-,=[O])([#6])[OX2][!#1]") is False):
                        fix_atom = True

                    # Alert for unexpected atoms bound to the neighboring nitrogen/oxygen.
                    elif atm_obj.matches_smarts("[C;X4,X3](-,=[N,O])([#6])[OX2][!#1]"):
                        logger.debug("While checking for inconsistencies in the atom CD of the residue %s, it was found an unexpected "
                                     "atom covalently bound to a neighboring nitrogen/oxygen. So, CD will not be amended."
                                     % pdb_atm.parent)

                    if fix_atom:
                        bond_types = []
                        for bond_obj in atm_obj.get_bonds():
                            partner_obj = bond_obj.get_partner_atom(atm_obj)

                            if partner_obj.get_idx() in pdb_map:
                                # Redundant rules: OE1 and OE2 section already fixes the bonds with CD.
                                if partner_obj.get_idx() in pdb_map and pdb_map[partner_obj.get_idx()].name == "OE1":
                                    # If the OE1 atom is bound to the below metals, update the CD-OE1 bond to CD=OE1.
                                    if partner_obj.matches_smarts(f"[OX2]({METAL_ATOM})"):
                                        bond_types.append((bond_obj, BondType.DOUBLE))
                                    else:
                                        bond_types.append((bond_obj, BondType.SINGLE))

                                # Redundant rules: OE1 and OE2 section already fixes the bonds with CD.
                                elif partner_obj.get_idx() in pdb_map and pdb_map[partner_obj.get_idx()].name == "OE2":
                                    # If the second oxygen (OE1) is bound to some atom not comprised by the expected metals, but OE2 is,
                                    # then update the CD-OE2 bond to CD=OE2.
                                    if partner_obj.matches_smarts(f"[OX2]({METAL_ATOM})[CX4]([#6])"
                                                                  f"[OX2;$(O([CX4])[!#1]);!$(O{METAL_ATOM})]"):
                                        bond_types.append((bond_obj, BondType.DOUBLE))
                                    else:
                                        bond_types.append((bond_obj, BondType.SINGLE))

                                # Any other atom bound to CD
                                else:
                                    bond_types.append((bond_obj, BondType.SINGLE))

                        self._fix_atom(atm_obj, bond_types=bond_types, charge=0, implicit_h_count=0)

            # Atom: CD1
            # Sanity check for all CD1 carbons with invalid bond types.
            elif pdb_atm.name == "CD1":

                if (pdb_atm.parent.resname in ["ILE", "LEU"] and atm_obj.get_neighbors_number(True) == 1 and
                        (atm_obj.get_valence() != 4 or atm_obj.get_charge() != 0 or atm_obj.has_only_bond_type(BondType.SINGLE) is False or
                            atm_obj.get_h_count() != 3 or atm_obj.is_in_ring() or atm_obj.is_aromatic())):

                    self._fix_atom(atm_obj, bond_types=[BondType.SINGLE], charge=0, implicit_h_count=3)

                elif (pdb_atm.parent.resname in ["PHE", "TRP", "TYR"] and atm_obj.get_neighbors_number(True) == 2 and
                        (atm_obj.get_valence() != 4 or atm_obj.get_charge() != 0 or atm_obj.get_degree() != 3 or
                            atm_obj.get_h_count() != 1 or atm_obj.is_aromatic() is False)):

                    bond_types = []
                    for bond_obj in atm_obj.get_bonds():
                        partner_obj = bond_obj.get_partner_atom(atm_obj)
                        if partner_obj.get_idx() in pdb_map:
                            # Double bonds with CG
                            if pdb_map[partner_obj.get_idx()].name == "CG":
                                bond_types.append((bond_obj, BondType.DOUBLE, True))
                            # Single bonds with CE1 and NE1
                            else:
                                bond_types.append((bond_obj, BondType.SINGLE, True))

                    self._fix_atom(atm_obj, bond_types=bond_types, charge=0, implicit_h_count=1, is_aromatic=True)

            # Atom: CD2
            # Sanity check for all CD2 carbons with invalid bond types.
            elif pdb_atm.name == "CD2":

                if (pdb_atm.parent.resname == "LEU" and atm_obj.get_neighbors_number(True) == 1 and
                        (atm_obj.get_valence() != 4 or atm_obj.get_charge() != 0 or atm_obj.has_only_bond_type(BondType.SINGLE) is False or
                            atm_obj.get_h_count() != 3 or atm_obj.is_in_ring() or atm_obj.is_aromatic())):

                    self._fix_atom(atm_obj, bond_types=[BondType.SINGLE], charge=0, implicit_h_count=3)

                elif (pdb_atm.parent.resname in ["HIS", "PHE", "TYR"] and atm_obj.get_neighbors_number(True) == 2 and
                        (atm_obj.get_valence() != 4 or atm_obj.get_charge() != 0 or atm_obj.get_degree() != 3 or
                            atm_obj.get_h_count() != 1 or atm_obj.is_aromatic() is False)):

                    bond_types = []
                    for bond_obj in atm_obj.get_bonds():
                        partner_obj = bond_obj.get_partner_atom(atm_obj)
                        if partner_obj.get_idx() in pdb_map:
                            # Bonds with CG
                            if pdb_map[partner_obj.get_idx()].name == "CG":
                                # Double bond with CG (only in HIS)
                                if pdb_atm.parent.resname == "HIS":
                                    bond_types.append((bond_obj, BondType.DOUBLE, True))
                                # Single bonds with CG (all, except HIS)
                                else:
                                    bond_types.append((bond_obj, BondType.SINGLE, True))
                            # Double bonds with CE2 (all, except HIS)
                            elif pdb_map[partner_obj.get_idx()].name == "CE2":
                                bond_types.append((bond_obj, BondType.DOUBLE, True))
                            # Single bonds with NE2 (only HIS)
                            elif pdb_map[partner_obj.get_idx()].name == "NE2":
                                bond_types.append((bond_obj, BondType.SINGLE, True))

                    self._fix_atom(atm_obj, bond_types=bond_types, charge=0, implicit_h_count=1, is_aromatic=True)

                elif (pdb_atm.parent.resname == "TRP" and atm_obj.get_neighbors_number(True) == 3 and
                        (atm_obj.get_valence() != 4 or atm_obj.get_charge() != 0 or atm_obj.get_degree() != 3 or
                            atm_obj.get_h_count() != 0 or atm_obj.is_aromatic() is False)):

                    bond_types = []
                    for bond_obj in atm_obj.get_bonds():
                        partner_obj = bond_obj.get_partner_atom(atm_obj)
                        if partner_obj.get_idx() in pdb_map:
                            # Double bonds with CE2
                            if pdb_map[partner_obj.get_idx()].name == "CE2":
                                bond_types.append((bond_obj, BondType.DOUBLE, True))
                            # Single bonds with CG and CE3
                            else:
                                bond_types.append((bond_obj, BondType.SINGLE, True))

                    self._fix_atom(atm_obj, bond_types=bond_types, charge=0, implicit_h_count=0, is_aromatic=True)

            # Atom: CE
            # Sanity check for all CE carbons with invalid bond types.
            elif pdb_atm.name == "CE":

                # CE from LYS.
                if (pdb_atm.parent.resname == "LYS" and atm_obj.get_neighbors_number(True) == 2 and
                        (atm_obj.get_valence() != 4 or atm_obj.get_charge() != 0 or atm_obj.has_only_bond_type(BondType.SINGLE) is False or
                            atm_obj.get_h_count() != 2 or atm_obj.is_in_ring() or atm_obj.is_aromatic())):

                    self._fix_atom(atm_obj, bond_types=[BondType.SINGLE], charge=0, implicit_h_count=2)

                # CE from MET.
                elif (pdb_atm.parent.resname == "MET" and atm_obj.get_neighbors_number(True) == 1 and
                        (atm_obj.get_valence() != 4 or atm_obj.get_charge() != 0 or atm_obj.has_only_bond_type(BondType.SINGLE) is False or
                            atm_obj.get_h_count() != 3 or atm_obj.is_in_ring() or atm_obj.is_aromatic())):

                    self._fix_atom(atm_obj, bond_types=[BondType.SINGLE], charge=0, implicit_h_count=3)

            # Atom: CE1
            # Sanity check for all CE1 carbons with invalid bond types.
            elif pdb_atm.name == "CE1":

                if (pdb_atm.parent.resname in ["HIS", "PHE", "TYR"] and atm_obj.get_neighbors_number(True) == 2 and
                        (atm_obj.get_valence() != 4 or atm_obj.get_charge() != 0 or atm_obj.get_degree() != 3 or
                            atm_obj.get_h_count() != 1 or atm_obj.is_aromatic() is False)):

                    bond_types = []
                    for bond_obj in atm_obj.get_bonds():
                        partner_obj = bond_obj.get_partner_atom(atm_obj)
                        if partner_obj.get_idx() in pdb_map:
                            # Single bonds with CD1 (all, except HYS)
                            if pdb_map[partner_obj.get_idx()].name == "CD1":
                                bond_types.append((bond_obj, BondType.SINGLE, True))
                            # Double bonds with CZ (all, except HYS)
                            elif pdb_map[partner_obj.get_idx()].name == "CZ":
                                bond_types.append((bond_obj, BondType.DOUBLE, True))
                            # Double bonds with ND1 (only HYS)
                            elif pdb_map[partner_obj.get_idx()].name == "ND1":
                                bond_types.append((bond_obj, BondType.DOUBLE, True))
                            # Single bonds with NE2 (only HYS)
                            elif pdb_map[partner_obj.get_idx()].name == "NE2":
                                # CE1 - NE2
                                bond_types.append((bond_obj, BondType.SINGLE, True))

                    self._fix_atom(atm_obj, bond_types=bond_types, charge=0, implicit_h_count=1, is_aromatic=True)

            # Atom: CE2
            # Sanity check for all CE2 carbons with invalid bond types.
            elif pdb_atm.name == "CE2":

                if (pdb_atm.parent.resname in ["PHE", "TYR"] and atm_obj.get_neighbors_number(True) == 2 and
                        (atm_obj.get_valence() != 4 or atm_obj.get_charge() != 0 or atm_obj.get_degree() != 3 or
                            atm_obj.get_h_count() != 1 or atm_obj.is_aromatic() is False)):

                    bond_types = []
                    for bond_obj in atm_obj.get_bonds():
                        partner_obj = bond_obj.get_partner_atom(atm_obj)
                        if partner_obj.get_idx() in pdb_map:
                            # Double bonds with CD2
                            if pdb_map[partner_obj.get_idx()].name == "CD2":
                                bond_types.append((bond_obj, BondType.DOUBLE, True))
                            # Single bonds with CZ
                            elif pdb_map[partner_obj.get_idx()].name == "CZ":
                                bond_types.append((bond_obj, BondType.SINGLE, True))

                    self._fix_atom(atm_obj, bond_types=bond_types, charge=0, implicit_h_count=1, is_aromatic=True)

                elif (pdb_atm.parent.resname == "TRP" and atm_obj.get_neighbors_number(True) == 3 and
                        (atm_obj.get_valence() != 4 or atm_obj.get_charge() != 0 or atm_obj.get_degree() != 3 or
                            atm_obj.get_h_count() != 0 or atm_obj.is_aromatic() is False)):

                    bond_types = []
                    for bond_obj in atm_obj.get_bonds():
                        partner_obj = bond_obj.get_partner_atom(atm_obj)
                        if partner_obj.get_idx() in pdb_map:
                            # Double bond with CD2
                            if pdb_map[partner_obj.get_idx()].name == "CD2":
                                bond_types.append((bond_obj, BondType.DOUBLE, True))
                            # Single bonds with NE1 and CZ2
                            else:
                                bond_types.append((bond_obj, BondType.SINGLE, True))

                    self._fix_atom(atm_obj, bond_types=bond_types, charge=0, implicit_h_count=0, is_aromatic=True)

            # Atom: CZ
            # Sanity check for all CZ carbons with invalid bond types.
            elif pdb_atm.name == "CZ":

                if (pdb_atm.parent.resname == "PHE" and atm_obj.get_neighbors_number(True) == 2 and
                        (atm_obj.get_valence() != 4 or atm_obj.get_charge() != 0 or atm_obj.get_degree() != 3 or
                            atm_obj.get_h_count() != 1 or atm_obj.is_aromatic() is False)):

                    bond_types = []
                    for bond_obj in atm_obj.get_bonds():
                        partner_obj = bond_obj.get_partner_atom(atm_obj)
                        if partner_obj.get_idx() in pdb_map:
                            # Double bond with CE1
                            if pdb_map[partner_obj.get_idx()].name == "CE1":
                                bond_types.append((bond_obj, BondType.DOUBLE, True))
                            # Single bond with CE2
                            else:
                                bond_types.append((bond_obj, BondType.SINGLE, True))

                    self._fix_atom(atm_obj, bond_types=bond_types, charge=0, implicit_h_count=1, is_aromatic=True)

                elif (pdb_atm.parent.resname == "TYR" and atm_obj.get_neighbors_number(True) == 3 and
                        (atm_obj.get_valence() != 4 or atm_obj.get_charge() != 0 or atm_obj.get_degree() != 3 or
                            atm_obj.get_h_count() != 0 or atm_obj.is_aromatic() is False)):

                    bond_types = []
                    for bond_obj in atm_obj.get_bonds():
                        partner_obj = bond_obj.get_partner_atom(atm_obj)
                        if partner_obj.get_idx() in pdb_map:
                            # Double bond with CE1
                            if pdb_map[partner_obj.get_idx()].name == "CE1":
                                bond_types.append((bond_obj, BondType.DOUBLE, True))
                            # Single bond with CE2
                            elif pdb_map[partner_obj.get_idx()].name == "CE2":
                                bond_types.append((bond_obj, BondType.SINGLE, True))
                            # Single bond with OH
                            else:
                                bond_types.append((bond_obj, BondType.SINGLE))

                    self._fix_atom(atm_obj, bond_types=bond_types, charge=0, implicit_h_count=0, is_aromatic=True)

                elif (pdb_atm.parent.resname == "ARG" and atm_obj.get_neighbors_number(True) == 3 and
                        (atm_obj.get_valence() != 4 or atm_obj.get_charge() != 0 or atm_obj.get_degree() != 3 or
                            atm_obj.get_h_count() != 0 or atm_obj.is_in_ring() or atm_obj.is_aromatic())):

                    bond_types = []
                    for bond_obj in atm_obj.get_bonds():
                        partner_obj = bond_obj.get_partner_atom(atm_obj)
                        if partner_obj.get_idx() in pdb_map:
                            # Double bond with NH2
                            if pdb_map[partner_obj.get_idx()].name == "NH2":
                                bond_types.append((bond_obj, BondType.DOUBLE))
                            # Single bonds with NE and NH1
                            else:
                                bond_types.append((bond_obj, BondType.SINGLE))

                    self._fix_atom(atm_obj, bond_types=bond_types, charge=0, implicit_h_count=0)

            # Atom: CE3, CZ2, CZ3, CH2
            # Sanity check for TRP carbons with invalid bond types.
            elif (pdb_atm.name in ["CE3", "CZ2", "CZ3", "CH2"] and atm_obj.get_neighbors_number(True) == 2 and
                    (atm_obj.get_valence() != 4 or atm_obj.get_charge() != 0 or atm_obj.get_degree() != 3 or
                        atm_obj.get_h_count() != 1 or atm_obj.is_aromatic() is False)):

                bond_types = []
                for bond_obj in atm_obj.get_bonds():
                    partner_obj = bond_obj.get_partner_atom(atm_obj)
                    if partner_obj.get_idx() in pdb_map:

                        # Double bonds
                        if ((pdb_atm.name == "CE3" and pdb_map[partner_obj.get_idx()].name == "CZ3") or
                                (pdb_atm.name == "CZ3" and pdb_map[partner_obj.get_idx()].name == "CE3") or
                                (pdb_atm.name == "CZ2" and pdb_map[partner_obj.get_idx()].name == "CH2") or
                                (pdb_atm.name == "CH2" and pdb_map[partner_obj.get_idx()].name == "CZ2")):

                            bond_types.append((bond_obj, BondType.DOUBLE, True))
                        # Single bonds
                        else:
                            bond_types.append((bond_obj, BondType.SINGLE, True))

                self._fix_atom(atm_obj, bond_types=bond_types, charge=0, implicit_h_count=1, is_aromatic=True)

            # Atom: ND1
            # Sanity check for HIS:ND1 with invalid bond types.
            elif pdb_atm.name == "ND1":

                fix_atom = False
                # HIS with metallic coordination perceived as covalent bond.
                #
                # Sometimes, the aromatic ring in HIS becomes a simple ring, therefore we should check for any
                # type of nitrogen in a ring, i.e, aromatic or aliphatic. Note that the aromatic ring
                # will be fixed after removing the bond with the metal.
                #
                # E.g.: 1USN:A:HIS:179
                #       1IUZ:A:HIS:37
                #
                if atm_obj.matches_smarts(f"[#7;R]-{METAL_ATOM}"):
                    # If it is necessary to break covalent bonds with metals.
                    if self.break_metal_bonds:
                        self._remove_metallic_bond(atm_obj)
                        fix_atom = True
                    else:
                        logger.debug("While checking for inconsistencies in the atom ND1 of the residue %s, it was found a metal "
                                     "covalently bound to it. However, nothing will be done because 'break_metal_bonds' "
                                     "was set to False." % pdb_atm.parent)

                # If the nitrogen is bound to something else not comprised in the previous rule, it is better not to
                # update anything. Otherwise, fix ND1.
                elif (atm_obj.matches_smarts("[#7;R]([#6])([#6])[!#1]") is False and
                      atm_obj.get_neighbors_number(True) == 2 and (atm_obj.get_valence() != 3 or atm_obj.get_charge() != 0 or
                                                                   atm_obj.get_degree() != 2 or atm_obj.get_h_count() != 0 or
                                                                   atm_obj.is_aromatic() is False)):
                    fix_atom = True

                # Alert for unexpected atoms bound to the nitrogen ND1.
                elif atm_obj.matches_smarts("[#7;R]([#6])([#6])[!#1]"):
                    logger.debug("While checking for inconsistencies in the atom ND1 of the residue %s, it was found an unexpected "
                                 "atom covalently bound to it. So, ND1 will not be amended." % pdb_atm.parent)

                if fix_atom:
                    bond_types = []
                    for bond_obj in atm_obj.get_bonds():
                        partner_obj = bond_obj.get_partner_atom(atm_obj)
                        if partner_obj.get_idx() in pdb_map:
                            # Double bond with CE1
                            if pdb_map[partner_obj.get_idx()].name == "CE1":
                                bond_types.append((bond_obj, BondType.DOUBLE, True))
                            # Single bonds with CG
                            else:
                                bond_types.append((bond_obj, BondType.SINGLE, True))

                    self._fix_atom(atm_obj, bond_types=bond_types, charge=0, implicit_h_count=0, is_aromatic=True)

            elif (pdb_atm.name == "ND1" and atm_obj.get_neighbors_number(True) == 2 and
                    (atm_obj.get_valence() != 3 or atm_obj.get_charge() != 0 or atm_obj.get_degree() != 2 or
                        atm_obj.get_h_count() != 0 or atm_obj.is_aromatic() is False)):

                bond_types = []
                for bond_obj in atm_obj.get_bonds():
                    partner_obj = bond_obj.get_partner_atom(atm_obj)
                    if partner_obj.get_idx() in pdb_map:
                        # Double bond with CE1
                        if pdb_map[partner_obj.get_idx()].name == "CE1":
                            bond_types.append((bond_obj, BondType.DOUBLE, True))
                        # Single bonds with CG
                        else:
                            bond_types.append((bond_obj, BondType.SINGLE, True))

                self._fix_atom(atm_obj, bond_types=bond_types, charge=0, implicit_h_count=0, is_aromatic=True)

            # Atom: ND2
            # Sanity check for ASN:ND2 with invalid bond types.
            elif (pdb_atm.name == "ND2" and atm_obj.get_neighbors_number(True) == 1 and
                    (atm_obj.get_valence() != 3 or atm_obj.get_charge() != 0 or atm_obj.has_only_bond_type(BondType.SINGLE) is False or
                        atm_obj.get_h_count() != 2 or atm_obj.is_in_ring() or atm_obj.is_aromatic())):

                # ASN with metallic coordination perceived as covalent bond.
                #
                #   It identifies two forms:
                #       - the firt consists of the usual form generated by Open Babel, where the double bond between
                #           OD1 and CG becomes a single bond, and it adds an additional single bond with a Metal.
                #
                #       - the second form may appear after correcting CG first, so the missing double bond would have already been
                #           fixed.
                #
                #   E.g.: 4FVR:A:ASN:678
                #
                if atm_obj.matches_smarts(f"N[C;X4,X3]([#6])-,=[OX2]{METAL_ATOM}"):
                    # If it is necessary to break covalent bonds with metals.
                    if self.break_metal_bonds:
                        self._fix_atom(atm_obj, bond_types=[BondType.SINGLE], charge=0, implicit_h_count=2)
                    else:
                        logger.debug("While checking for inconsistencies in the atom ND2 of the residue %s, it was found a metal "
                                     "covalently bound to the atom OD1. However, nothing will be done because 'break_metal_bonds' "
                                     "was set to False." % pdb_atm.parent)

                # If the oxygen is bound to something else not comprised in the previous rules, it is better not to update anything.
                # Otherwise, fix the ND2.
                elif atm_obj.matches_smarts("N[CX4]([#6])[OX2][!#1]") is False:

                    self._fix_atom(atm_obj, bond_types=[BondType.SINGLE], charge=0, implicit_h_count=2)

                # Alert for unexpected atoms bound to the oxygen OD1.
                elif atm_obj.matches_smarts("N[CX4]([#6])[OX2][!#1]"):
                    logger.debug("While checking for inconsistencies in the atom ND2 of the residue %s, it was found an unexpected "
                                 "atom covalently bound to the oxygen OD1. So, ND2 will not be amended." % pdb_atm.parent)

            # Atom: NE
            # Sanity check for ARG:NE with invalid bond types.
            elif (pdb_atm.name == "NE" and atm_obj.get_neighbors_number(True) == 2 and
                    (atm_obj.get_valence() != 3 or atm_obj.get_charge() != 0 or atm_obj.has_only_bond_type(BondType.SINGLE) is False or
                        atm_obj.get_h_count() != 1 or atm_obj.is_in_ring() or atm_obj.is_aromatic())):

                self._fix_atom(atm_obj, bond_types=[BondType.SINGLE], charge=0, implicit_h_count=1)

            # Atom: NE1
            # Sanity check for TRP:NE1 with invalid bond types.
            elif (pdb_atm.name == "NE1" and atm_obj.get_neighbors_number(True) == 2 and
                    (atm_obj.get_valence() != 3 or atm_obj.get_charge() != 0 or atm_obj.get_degree() != 3 or
                        atm_obj.get_h_count() != 1 or atm_obj.is_aromatic() is False)):

                bond_types = [(bond_obj, BondType.SINGLE, True) for bond_obj in atm_obj.get_bonds()]
                self._fix_atom(atm_obj, bond_types=bond_types, charge=0, implicit_h_count=1, is_aromatic=True)

            # Atom: NE2
            # Sanity check for NE2 nitrogens with invalid bond types.
            elif pdb_atm.name == "NE2":

                if pdb_atm.parent.resname == "HIS":
                    # HIS with metallic coordination perceived as covalent bond.
                    #
                    # Sometimes, the aromatic ring in HIS becomes a simple ring, therefore we check for any
                    # type of nitrogen in a ring, i.e, aromatic or aliphatic. Note that the aromatic ring
                    # will be fixed after removing the bond with the metal.
                    #
                    # E.g.: 1USN:A:HIS:151
                    #       1USN:A:HIS:166
                    if atm_obj.matches_smarts(f"[#7;R]{METAL_ATOM}"):
                        # If it is necessary to break covalent bonds with metals.
                        if self.break_metal_bonds:
                            self._remove_metallic_bond(atm_obj)

                            bond_types = [(bond_obj, BondType.SINGLE, True) for bond_obj in atm_obj.get_bonds()]
                            self._fix_atom(atm_obj, bond_types=bond_types, charge=0, implicit_h_count=1, is_aromatic=True)
                        else:
                            logger.debug("While checking for inconsistencies in the atom NE2 of the residue %s, it was found a metal "
                                         "covalently bound to it. However, nothing will be done because 'break_metal_bonds' "
                                         "was set to False." % pdb_atm.parent)

                    # If the nitrogen is bound to something else not comprised in the previous rule, it is better not to
                    # update anything. Otherwise, fix NE2.
                    elif (atm_obj.matches_smarts("[#7;R]([#6])([#6])[!#1]") is False and
                          atm_obj.get_neighbors_number(True) == 2 and (atm_obj.get_valence() != 3 or atm_obj.get_charge() != 0 or
                                                                       atm_obj.get_degree() != 3 or atm_obj.get_h_count() != 1 or
                                                                       atm_obj.is_aromatic() is False)):

                        bond_types = [(bond_obj, BondType.SINGLE, True) for bond_obj in atm_obj.get_bonds()]
                        self._fix_atom(atm_obj, bond_types=bond_types, charge=0, implicit_h_count=1, is_aromatic=True)

                    # Alert for unexpected atoms bound to the nitrogen NE2.
                    elif atm_obj.matches_smarts("[#7;R]([#6])([#6])[!#1]"):
                        logger.debug("While checking for inconsistencies in the atom NE2 of the residue %s, it was found an unexpected "
                                     "atom covalently bound to it. So, NE2 will not be amended." % pdb_atm.parent)

                elif (pdb_atm.parent.resname == "GLN" and atm_obj.get_neighbors_number(True) == 1 and
                        (atm_obj.get_valence() != 3 or atm_obj.get_charge() != 0 or atm_obj.has_only_bond_type(BondType.SINGLE) is False or
                            atm_obj.get_h_count() != 2 or atm_obj.is_in_ring() or atm_obj.is_aromatic())):

                    # GLN with metallic coordination perceived as covalent bond.
                    #
                    #   It identifies two forms:
                    #       - the firt consists of the usual form generated by Open Babel, where the double bond between
                    #           OE1 and CD becomes a single bond, and it adds an additional single bond with a Metal.
                    #
                    #       - the second form may appear after correcting CD first, so the missing double bond would have already been
                    #           fixed.
                    #
                    if atm_obj.matches_smarts(f"N[C;X4,X3]([#6])-,=[OX2]{METAL_ATOM}"):
                        # If it is necessary to break covalent bonds with metals.
                        if self.break_metal_bonds:
                            self._fix_atom(atm_obj, bond_types=[BondType.SINGLE], charge=0, implicit_h_count=2)
                        else:
                            logger.debug("While checking for inconsistencies in the atom NE2 of the residue %s, it was found a metal "
                                         "covalently bound to the atom OE1. However, nothing will be done because 'break_metal_bonds' "
                                         "was set to False." % pdb_atm.parent)

                    # If the oxygen is bound to something else not comprised in the previous rules, it is better not to update anything.
                    # Otherwise, fix the NE2.
                    elif atm_obj.matches_smarts("N[CX4]([#6])[OX2][!#1]") is False:
                        self._fix_atom(atm_obj, bond_types=[BondType.SINGLE], charge=0, implicit_h_count=2)

                    # Alert for unexpected atoms bound to the oxygen OE1.
                    elif atm_obj.matches_smarts("N[CX4]([#6])[OX2][!#1]"):
                        logger.debug("While checking for inconsistencies in the atom NE2 of the residue %s, it was found an unexpected "
                                     "atom covalently bound to the oxygen OE1. So, NE2 will not be amended." % pdb_atm.parent)

            # Atom: NZ
            # Sanity check for LYS:NZ with invalid bond types.
            elif (pdb_atm.name == "NZ" and atm_obj.get_neighbors_number(True) == 1 and
                    (atm_obj.get_valence() != 4 or atm_obj.get_charge() != 1 or atm_obj.get_degree() != 4 or
                        atm_obj.get_h_count() != 3 or atm_obj.is_in_ring() or atm_obj.is_aromatic())):

                self._fix_atom(atm_obj, bond_types=[BondType.SINGLE], charge=1, implicit_h_count=3)

            # Atom: NH1
            # Sanity check for ARG:NH1 with invalid bond types.
            elif (pdb_atm.name == "NH1" and atm_obj.get_neighbors_number(True) == 1 and
                    (atm_obj.get_valence() != 3 or atm_obj.get_charge() != 0 or atm_obj.get_degree() != 3 or
                        atm_obj.get_h_count() != 2 or atm_obj.is_in_ring() or atm_obj.is_aromatic())):

                self._fix_atom(atm_obj, bond_types=[BondType.SINGLE], charge=0, implicit_h_count=2)

            # Atom: NH2
            # Sanity check for ARG:NH2 with invalid bond types.
            elif (pdb_atm.name == "NH2" and atm_obj.get_neighbors_number(True) == 1 and
                    (atm_obj.get_valence() != 4 or atm_obj.get_charge() != 1 or atm_obj.get_degree() != 3 or
                        atm_obj.get_h_count() != 2 or atm_obj.is_in_ring() or atm_obj.is_aromatic())):

                self._fix_atom(atm_obj, bond_types=[BondType.DOUBLE], charge=1, implicit_h_count=2)

            # Atom: O, OD1, and OE1
            # Sanity check for OD1/OE1 oxygens with invalid bond types.
            elif pdb_atm.name in ["O", "OD1", "OE1"]:

                # Any O oxygen or ASN/ASP/GLN/GLU OD1/OE1 with metallic coordination perceived as covalent bond.
                #
                #   It identifies two forms:
                #       - the firt consists of the usual form generated by Open Babel where the double bonds between O and C,
                #           OD1 and CG, or OE1 and CD become single bonds, while it adds an additional single bond with a Metal.
                #
                #       - the second form may appear after correcting C/CG/CD first, so the missing double bond would have already been
                #           fixed, but it makes the oxygens to have incorrect valence and degree, so we still need to fix them.
                #           And of course, we still need to remove the bond with the metal.
                #
                # E.g.: 6JWU:A:GLU:42, 6JWU:B:ASP:210, 4FVR:A:ASN:678.
                #
                if (atm_obj.matches_smarts(f"[OX2]({METAL_ATOM})-,=[C;X4,X3]([#6])(-,=[N])") or
                        atm_obj.matches_smarts(f"[OX2]({METAL_ATOM})-,=[C;X4,X3]([#6])(-,=[O])")):
                    # If it is necessary to break covalent bonds with metals.
                    if self.break_metal_bonds:
                        self._remove_metallic_bond(atm_obj)
                        self._fix_atom(atm_obj, bond_types=[BondType.DOUBLE], charge=0, implicit_h_count=0)
                    else:
                        logger.debug("While checking for inconsistencies in the atom %s of the residue %s, it was found a metal "
                                     "covalently bound to it. However, nothing will be done because 'break_metal_bonds' "
                                     "was set to False." % (pdb_atm.name, pdb_atm.parent))

                # If the oxygen is bound to something else not comprised in the previous rules, it is better not to update anything.
                # Otherwise, fix the oxygen.
                #
                # Note that it doesn't check out the situation of the other oxygen (OXT or OD2/OE2 in ASP/GLU), because O/OD1/OE1 will
                # always have a double bond no matter the other oxygen has or not a bond with metals or other atoms.
                #
                # E.g.: 3QQL:A:GLY:11 contain incorrectly perceived bonds.
                #
                elif (atm_obj.get_neighbors_number(True) == 1 and (atm_obj.get_valence() != 2 or atm_obj.get_charge() != 0 or
                                                                   atm_obj.has_only_bond_type(BondType.DOUBLE) is False or
                                                                   atm_obj.get_h_count() != 0 or atm_obj.is_in_ring() or
                                                                   atm_obj.is_aromatic())):

                    self._fix_atom(atm_obj, bond_types=[BondType.DOUBLE], charge=0, implicit_h_count=0)

                # Alert for unexpected atoms bound to the oxygens.
                elif atm_obj.matches_smarts(f"[OX2;$(O([C;X4,X3])[!#1]);!$(O{METAL_ATOM})]-,=[C;X4,X3]"):
                    logger.debug("While checking for inconsistencies in the atom %s of the residue %s, it was found an unexpected "
                                 "atom covalently bound to it. So, %s will not be amended." % (pdb_atm.name, pdb_atm.parent,
                                                                                                 pdb_atm.name))

            # Atom: OXT, OD2, and OE2
            # Sanity check for OXT/OD2/OE2 oxygens with invalid bond types.
            elif pdb_atm.name in ["OXT", "OD2", "OE2"]:

                # ASP/GLU or any OXT with metallic coordination perceived as covalent bond.
                #
                #   It considers monodentate (only OXT/OD2/E2) and bidentate (both oxygens in ASP/GLU or main chain O and OXT)
                #   interactions with metals.
                #
                #   For the second oxygen (O/OD1/OE1), we also consider that it may appear after correcting C/CG/CD first, so the
                #   missing double bond would have already been fixed.
                #
                #   E.g.: 6JWU:B:ASP:210.
                #
                if (atm_obj.matches_smarts(f"[OX2]({METAL_ATOM})[C;X4,X3]([#6])-,=[OX2;$(O{METAL_ATOM})]") or
                        atm_obj.matches_smarts(f"[OX2]({METAL_ATOM})[CX3]([#6])=[OX1]")):
                    # If it is necessary to break covalent bonds with metals.
                    if self.break_metal_bonds:
                        self._remove_metallic_bond(atm_obj)
                        self._fix_atom(atm_obj, bond_types=[BondType.SINGLE], charge=-1, implicit_h_count=0)
                    else:
                        logger.debug("While checking for inconsistencies in the atom %s of the residue %s, it was found a metal "
                                     "covalently bound to it. However, nothing will be done because 'break_metal_bonds' "
                                     "was set to False." % (pdb_atm.name, pdb_atm.parent))

                # If the second oxygen (OD1/OE1) is bound to something else not comprised in the previous rules,
                # then it won't be fixed and, therefore, we should update OD2/OE2 as follows: remove the bond with
                # the metal and then substitute the single bond with CG/CD for a double bond. Note that OD2/OE2
                # will play the role of OD1/OE1 after the update because that oxygen won't be amended.
                #
                # Note, we also consider that it may appear after correcting C/CG/CD first, so the missing double
                # bond would have already been fixed.
                elif atm_obj.matches_smarts(f"[OX2]({METAL_ATOM})-,=[C;X4,X3]([#6])[OX2;$(O([C;X4,X3])[!#1])]"):
                    # If it is necessary to break covalent bonds with metals.
                    if self.break_metal_bonds:
                        self._remove_metallic_bond(atm_obj)
                        self._fix_atom(atm_obj, bond_types=[BondType.DOUBLE], charge=0, implicit_h_count=0)
                    else:
                        logger.debug("While checking for inconsistencies in the atom %s of the residue %s, it was found a metal "
                                     "covalently bound to it. However, nothing will be done because 'break_metal_bonds' "
                                     "was set to False." % (pdb_atm.name, pdb_atm.parent))

                # Capture cases where the second oxygen is bound to a metal atom, but not OXT/OD2/OE2.
                elif atm_obj.matches_smarts(f"[OX1]=[CX3]([#6])[OX2;$(O{METAL_ATOM})]"):
                    # If it is necessary to break covalent bonds with metals.
                    if self.break_metal_bonds:
                        self._fix_atom(atm_obj, bond_types=[BondType.SINGLE], charge=-1, implicit_h_count=0)
                    else:
                        logger.debug("While checking for inconsistencies in the atom %s of the residue %s, it was found a metal "
                                     "covalently bound to the other carboxyl oxygen. However, nothing will be done because "
                                     "'break_metal_bonds' was set to False." % (pdb_atm.name, pdb_atm.parent))

                # It fixes invalid oxygens, excluing cases where OD2/OE2 is bound to an atom not comprised in the
                # previous rules and cases where OD2/OE2 has a double bond with CG/CD and the second oxygen (OD1/OE1)
                # is bound to two atoms, one of which is an unexpected atom not considered in our standardization function
                elif (atm_obj.matches_smarts("[OX2]([!#1])[C;X4,X3]") is False and
                        atm_obj.matches_smarts(f"[OX1]=[CX3]([#6])[OX2;$(O([CX3])[!#1]);!$(O{METAL_ATOM})]") is False and
                        atm_obj.get_neighbors_number(True) == 1 and (atm_obj.get_valence() != 1 or atm_obj.get_charge() != -1 or
                                                                     atm_obj.get_degree() != 1 or atm_obj.get_h_count() != 0 or
                                                                     atm_obj.is_in_ring() or atm_obj.is_aromatic())):
                    self._fix_atom(atm_obj, bond_types=[BondType.SINGLE], charge=-1, implicit_h_count=0)

                # Alert for unexpected atoms bound to the oxygens.
                elif atm_obj.matches_smarts("[OX2]([!#1])[C;X4,X3]"):
                    logger.debug("While checking for inconsistencies in the atom %s of the residue %s, it was found an unexpected "
                                 "atom covalently bound to it. So, %s will not be amended." % (pdb_atm.name, pdb_atm.parent,
                                                                                                 pdb_atm.name))

                # Alert for unexpected atoms bound to the oxygens.
                elif atm_obj.matches_smarts(f"[OX1]=[CX3]([#6])[OX2;$(O([CX3])[!#1]);!$(O{METAL_ATOM})]"):
                    logger.debug("While checking for inconsistencies in the atom %s of the residue %s, it was found an unexpected "
                                 "atom covalently bound to the other carboxyl oxygen. So, %s was treated as a carbonyl oxygen." %
                                 (pdb_atm.name, pdb_atm.parent, pdb_atm.name))

            # Atom: OG, OG1, and OH
            # Sanity check for OG/OG1/OH oxygens with invalid bond types.
            elif pdb_atm.name in ["OG", "OG1", "OH"]:

                # TYR/SER/THR with metallic coordination perceived as covalent bond.
                #
                # Bertini et al. 2007. Biological Inorganic Chemistry: Structure and Reactivity.
                #
                # E.g: 1TFD:A:TYR:188
                #
                if atm_obj.matches_smarts(f"[OX2]{METAL_ATOM}"):
                    # If it is necessary to break covalent bonds with metals.
                    if self.break_metal_bonds:
                        self._remove_metallic_bond(atm_obj)
                        self._fix_atom(atm_obj, bond_types=[BondType.SINGLE], charge=0, implicit_h_count=1)
                    else:
                        logger.debug("While checking for inconsistencies in the atom %s of the residue %s, it was found a metal "
                                     "covalently bound to it. However, nothing will be done because 'break_metal_bonds' "
                                     "was set to False." % (pdb_atm.name, pdb_atm.parent))

                # If the oxygen is bound to something else not comprised in the previous rule, it is better not to
                # update anything. Otherwise, fix OG/OG1/OH.
                elif atm_obj.get_neighbors_number(True) == 1 and (atm_obj.get_valence() != 2 or atm_obj.get_charge() != 0 or
                                                                  atm_obj.has_only_bond_type(BondType.SINGLE) is False or
                                                                  atm_obj.get_h_count() != 1 or atm_obj.is_in_ring() or
                                                                  atm_obj.is_aromatic()):

                    self._fix_atom(atm_obj, bond_types=[BondType.SINGLE], charge=0, implicit_h_count=1)

                # Alert for unexpected atoms bound to the hydroxyl oxygens.
                elif atm_obj.matches_smarts(f"[OX2;$(O([#6])[!#1]);!$(O{METAL_ATOM})]"):
                    logger.debug("While checking for inconsistencies in the atom %s of the residue %s, it was found an unexpected "
                                 "atom covalently bound to it. So, %s will not be amended." % (pdb_atm.name, pdb_atm.parent,
                                                                                                 pdb_atm.name))

            # Atom: SD
            # Sanity check for MET:SD with invalid bond types.
            #
            # Although MET can be involved in the coordination of metals, due to its two single bonds, it will never be incorrectly
            # perceived as being covalently bound to metals. Therefore, we do not need to check for it.
            #
            elif (pdb_atm.name == "SD" and atm_obj.get_neighbors_number(True) == 2 and
                    (atm_obj.get_valence() != 2 or atm_obj.get_charge() != 0 or atm_obj.has_only_bond_type(BondType.SINGLE) is False or
                        atm_obj.get_h_count() != 0 or atm_obj.is_in_ring() or atm_obj.is_aromatic())):

                self._fix_atom(atm_obj, bond_types=[BondType.SINGLE], charge=0, implicit_h_count=0)

            # Atom: SG
            # Sanity check for CYS:SG with invalid bond types.
            elif pdb_atm.name == "SG":

                # CYS with metallic coordination perceived as covalent bond.
                #
                # Harding et al. 2010. Metals in protein structures: a review of their principal features.
                # Bertini et al. 2007. Biological Inorganic Chemistry: Structure and Reactivity.
                #
                # E.g: 1IUZ:A:CYS:84
                #
                if atm_obj.matches_smarts(f"[SX2]{METAL_ATOM}"):
                    # If it is necessary to break covalent bonds with metals.
                    if self.break_metal_bonds:
                        self._remove_metallic_bond(atm_obj)
                        self._fix_atom(atm_obj, bond_types=[BondType.SINGLE], charge=-1, implicit_h_count=0)
                    else:
                        logger.debug("While checking for inconsistencies in the atom SG of the residue %s, it was found a metal "
                                     "covalently bound to it. However, nothing will be done because 'break_metal_bonds' "
                                     "was set to False." % pdb_atm.parent)

                # If the sulfur is bound to something else not comprised in the previous rule, it is better not to
                # update anything. Otherwise, fix SG.
                elif (atm_obj.matches_smarts("[SX2](C)[!#1]") is False and
                        (atm_obj.get_valence() != 2 or atm_obj.get_charge() != 0 or atm_obj.get_degree() != 2 or
                         atm_obj.has_only_bond_type(BondType.SINGLE) is False or atm_obj.is_in_ring() or
                         atm_obj.is_aromatic())):

                    # Get the number of implicit hydrogen count based on the bonds SG has, i.e., if it is the usual cysteine, then
                    # it has 1 implicit hydrogen, but if it establishes a disulfide bond, then it has no hydrogen.
                    implicit_h_count = 0 if atm_obj.get_neighbors_number(True) == 2 else 1

                    self._fix_atom(atm_obj, bond_types=[BondType.SINGLE], charge=0, implicit_h_count=implicit_h_count)

                # Alert for unexpected atoms bound to the hydroxyl oxygens.
                elif atm_obj.matches_smarts("[SX2](C)[!#1]"):
                    logger.debug("While checking for inconsistencies in the atom SG of the residue %s, it was found an unexpected "
                                 "atom covalently bound to it. So, SG will not be amended." % pdb_atm.parent)

        # Only remove metals with no bond to any residue.
        for metal_obj in self.found_metals.values():
            if len(metal_obj.get_bonds()) == 0:
                self.removed_atoms.append(metal_obj.get_id())
                atm_obj.parent.unwrap().DeleteAtom(metal_obj)

    def _fix_atom(self, atm_obj, charge=None, remove_explict_h=True, implicit_h_count=None,
                  bond_types=None, is_aromatic=False, in_ring=False):

        # Remove all current explicit hydrogens.
        if remove_explict_h:
            self._remove_explict_hydrogens(atm_obj)

        # Convert bond types.
        if bond_types is not None:
            self._fix_bonds(atm_obj, bond_types)

        # All aromatic atoms will have their in_ring property set to True.
        if is_aromatic:
            in_ring = True

        # Set if atom belongs to a ring or not.
        atm_obj.set_in_ring(in_ring)

        # Set if an atom is aromatic or not.
        atm_obj.set_as_aromatic(is_aromatic)

        if implicit_h_count is not None:
            atm_obj.unwrap().SetImplicitHCount(implicit_h_count)

        # Set charge.
        if charge is not None:
            atm_obj.set_charge(charge)

    def _remove_explict_hydrogens(self, atm_obj):
        delete_hs = []
        for b in atm_obj.get_bonds():
            if b.get_partner_atom(atm_obj).get_symbol() == "H":
                delete_hs.append(b.get_partner_atom(atm_obj))

        for hs_obj in delete_hs:
            atm_obj.parent.unwrap().DeleteAtom(hs_obj.unwrap())

    def _fix_bonds(self, atm_obj, bond_types):
        # bond_types becomes an empty list if bond_types is None.
        bond_types = bond_types or []

        # If bond_types is an empty list, do nothing.
        if len(bond_types) == 0:
            return

        # All bonds will be converted to the same type.
        elif len(bond_types) == 1 and isinstance(bond_types[0], BondType):
            new_bond_type = bond_types[0]
            for bond_obj in atm_obj.get_bonds():
                bond_obj.set_bond_type(new_bond_type)

        # It expects a list of tuples where the first element is the bond to be updated, the second element should be the new type,
        # and the third optional element is a flag to indicate if the bond is aromatic or not.
        else:
            for bond_info in bond_types:
                bond_obj, new_bond_type = bond_info[0:2]
                # Update bond type
                bond_obj.set_bond_type(new_bond_type)

                # If the tuple contains three elements, it must be a boolean to define if it is an aromatic bond or not.
                if len(bond_info) == 3:
                    is_aromatic = bond_info[2]
                    bond_obj.set_as_aromatic(is_aromatic)

    def _remove_metallic_bond(self, atm_obj):

        bonds_to_remove = []
        for bond_obj in atm_obj.get_bonds():
            partner_obj = bond_obj.get_partner_atom(atm_obj)

            if partner_obj.get_symbol() in METALS:
                bonds_to_remove.append((partner_obj, bond_obj))

        # Remove metallic bond and mark metal to remotion.
        for metal_obj, bond_obj in bonds_to_remove:

            self.found_metals[metal_obj.get_id()] = metal_obj

            atm_obj.parent.unwrap().DeleteBond(bond_obj.unwrap())



import re


def get_atom_names_by_id(cif_file):
    """Read a single-molecule CIF file and return the molecule's atom names.

    In the current version, if applied on multi-molecular CIF files,
    only the first molecule's atom names are returned.

    Returns
    -------
     : dict
    """

    regex = re.compile(' {1,}')

    atom_names = {}
    with open(cif_file, "r") as IN:
        loop_read = False
        pdbx_ordinal_read = False

        multi_atom = False
        start_read = False

        last_line = -1

        for i, line in enumerate(IN.readlines()):
            line = line.strip()

            if line.startswith("loop_"):
                loop_read = True
            else:
                if line.startswith("_chem_comp_atom.comp_id"):
                    if loop_read and i == last_line + 1:
                        multi_atom = True
                    else:
                        start_read = True
                elif line.startswith("_chem_comp_atom.pdbx_ordinal"):
                    pdbx_ordinal_read = True
                elif pdbx_ordinal_read and line.startswith("#"):
                    break

                if start_read:
                    if multi_atom:
                        cols = regex.split(line)
                        atom_names[len(atom_names)] = cols[1]
                    elif line.startswith("_chem_comp_atom.atom_id"):
                        atom_names[len(atom_names)] = line[line.rfind(' ') + 1:]
                        break

            if pdbx_ordinal_read and multi_atom:
                start_read = True

            last_line = i

    return atom_names



import argparse
import os
import shutil

from luna import LocalProject
from luna.version import __version__ as version

from luna.mol.entry import MolFileEntry
from luna.interaction.filter import InteractionFilter, BindingModeFilter
from luna.interaction.calc import InteractionCalculator
from luna.interaction.view import InteractionViewer
from luna.interaction.fp.type import IFPType
from luna.util.file import get_filename, create_directory
from luna.interaction.calc import InteractionsManager
from luna.projects import EntryResults

IFP_TYPES = {
    "EIFP": IFPType.EIFP,
    "FIFP": IFPType.FIFP,
    "HIFP": IFPType.HIFP,
}


def get_entries(args):
    with open(args.entries_file, "r") as IN:
        entry_ids = set()
        for row in IN:
            row = row.strip()
            if row == "":
                continue
            entry_ids.add(row)

        pdb_id = get_filename(args.pdb_file)

        for ligand_id in entry_ids:
            yield MolFileEntry.from_mol_file(pdb_id, ligand_id, mol_file=args.ligand_file, is_multimol_file=True)


def main():

    parser = argparse.ArgumentParser()

    parser.add_argument('-p', "--prot", dest="pdb_file", type=str, required=True,
                        help="the protein PDB file")

    parser.add_argument('-e', "--entries", dest="entries_file", type=str, required=True,
                        help="an input file containing a list of ligand ids to process")

    parser.add_argument('-l', "--lig", dest="ligand_file", type=str, required=True,
                        help="a molecular file containing 1 or more ligands")

    parser.add_argument('-w', dest="working_path", type=str, required=True,
                        help="the path where the project and its results will be saved")

    parser.add_argument('--out_ifp', dest="out_ifp", action="store_true",
                        help="defines whether it should generate LUNA interaction fingerprints")

    parser.add_argument('-L', dest="ifp_num_levels", type=int, default=2,
                        help="the number of level defines the number of iterations to construct the fingerprint. "
                             "Default: 2")

    parser.add_argument('-R', dest="ifp_radius_step", type=float, default=5.73171,
                        help="the radius growth rate defines the multiplier to increase the sphere size at each level. "
                             "Default: 5.73171")

    parser.add_argument('-S', dest="ifp_length", type=int, default=4096,
                        help="the fingerprint length. Default: 4096")

    parser.add_argument('-T', dest="ifp_type", type=str, default="EIFP", choices=['EIFP', 'HIFP', 'FIFP'],
                        help="the fingerprint type. Default: EIFP")

    parser.add_argument('-B', dest="bit_ifp", action="store_true",
                        help="defines whether it should use bit fingerprints. The default value is False, "
                             "which imply that count fingerprints are used instead")

    parser.add_argument('-O', dest="ifp_output", type=str,
                        help="the fingerprint output file. Default: <WORKING_PATH>/results/fingerprints/ifp.csv")

    parser.add_argument('--sim_matrix_output', type=str, required=False,
                        help="the path where the similarity matrix will be saved. "
                             "If not provided, it won't be generated")

    parser.add_argument('--filter_binding_modes', dest="binding_modes_file", type=str,
                        help="the path of a file containing binding modes to filter")

    parser.add_argument('--out_pse', dest="out_pse", action="store_true",
                        help="defines whether it should export interactions to Pymol")

    parser.add_argument('--pse_path', dest="pse_path", type=str, required=False,
                        help="the path where Pymol sessions (PSE files) will be saved. "
                             "Default: <WORKING_PATH>/results/pse/")

    parser.add_argument('--overwrite', dest="overwrite", action="store_true",
                        help="defines whether it should overwrite an existing project")

    parser.add_argument('-f', '--fork_project', dest="fork_project", type=str, required=False,
                        help="If provided, copy an existing project to <WORKING_PATH>")

    parser.add_argument("--nproc", dest="nproc", type=int,
                        help="the number of processors to use")

    args = parser.parse_args()

    # Fork a project
    if args.fork_project:
        fork_proj_pkl_file = "%s/project_v%s.pkl.gz" % (args.fork_project, version)
        if not os.path.exists(fork_proj_pkl_file):
            raise OSError("No valid project was found at '%s'" % args.fork_project)
        shutil.copytree(args.fork_project, args.working_path)

    proj_pkl_file = "%s/project_v%s.pkl.gz" % (args.working_path, version)
    if not os.path.exists(proj_pkl_file) or args.overwrite:
        print(u"\u25a8 Creating a new LUNA project: '%s'...\n" % args.working_path)

        entries = list(get_entries(args))

        pdb_path = os.path.dirname(os.path.realpath(args.pdb_file))
        ic = InteractionCalculator(inter_filter=InteractionFilter.new_pli_filter(), strict_donor_rules=True, strict_weak_donor_rules=True)

        opts = {}
        opts["working_path"] = args.working_path
        opts["pdb_path"] = pdb_path
        opts["entries"] = entries
        opts["overwrite_path"] = args.overwrite
        opts["inter_calc"] = ic
        opts["mol_obj_type"] = 'rdkit'
        opts["add_h"] = True
        opts["amend_mol"] = True
        opts["calc_ifp"] = False
        opts["out_pse"] = False

        if args.nproc:
            opts["nproc"] = args.nproc

        pli_obj = LocalProject(**opts)
        pli_obj.run()
    else:
        print(u"\u25a8 Reloading existing project: '%s'...\n" % args.working_path)
        pli_obj = LocalProject.load(proj_pkl_file)
    print()

    if args.binding_modes_file:
        binding_mode_filter = BindingModeFilter.from_config_file(args.binding_modes_file)

        for agm in pli_obj.atm_grps_mngrs:
            im = InteractionsManager(agm.get_all_interactions(), agm.entry)
            im.filter_out_by_binding_mode(binding_mode_filter)

            entry_results = EntryResults(agm.entry, agm, im, None, None)
            pkl_file = "%s/chunks/%s.pkl.gz" % (pli_obj.working_path, agm.entry.to_string())
            entry_results.save(pkl_file)

            csv_file = "%s/results/interactions/%s.csv" % (pli_obj.working_path, agm.entry.to_string())
            im.to_csv(csv_file)

    if args.out_ifp:
        ifp_output = args.ifp_output
        if ifp_output is None:
            create_directory("%s/results/fingerprints" % args.working_path)
            ifp_output = "%s/results/fingerprints/ifp.csv" % args.working_path

        pli_obj.calc_ifp = True
        pli_obj.ifp_num_levels = args.ifp_num_levels
        pli_obj.ifp_radius_step = args.ifp_radius_step
        pli_obj.ifp_length = args.ifp_length
        pli_obj.ifp_type = IFP_TYPES[args.ifp_type]

        pli_obj.ifp_count = True
        if args.bit_ifp:
            pli_obj.ifp_count = False

        pli_obj.ifp_output = ifp_output

        if args.sim_matrix_output:
            pli_obj.ifp_sim_matrix_output = args.sim_matrix_output

        pli_obj.generate_ifps()

    if args.out_pse:
        pse_path = args.pse_path or f"{pli_obj.working_path}/results/pse"
        create_directory(pse_path)

        for ic in pli_obj.interactions_mngrs:
            pse_file = f"{pse_path}/%s.pse" % ic.entry.to_string()
            viewer = InteractionViewer(add_directional_arrows=False)
            viewer.new_session([(ic.entry, ic, pli_obj.pdb_path)], pse_file)


if __name__ == '__main__':
    print()
    main()



# Copyright (C) 2002, Thomas Hamelryck (thamelry@binf.ku.dk)
# This code is part of the Biopython distribution and governed by its
# license.  Please see the LICENSE file that should have been included
# as part of this package.

###################################################################

# Modifications included by Alexandre Fassio (alexandrefassio@dcc.ufmg.br).
# Date: 08/03/2018.

# 1) Inherit inhouse modifications. Package: MyBio.
# 2) StructureBuilder now parses CONECT records.
# 3) set_structure now creates a new Structure object with the same Id as the original structure.

# Each line or block with modifications contain a MODBY tag.

###################################################################

"""Output of PDB files."""

from Bio._py3k import basestring

# MODBY: Alexandre Fassio
# Inherit inhouse modifications. Package: MyBio.
from luna.MyBio.PDB.StructureBuilder import StructureBuilder  # To allow saving chains, residues, etc..
from Bio.Data.IUPACData import atom_weights  # Allowed Elements

_ATOM_FORMAT_STRING = "%s%5i %-4s%c%3s %c%4i%c   %8.3f%8.3f%8.3f%s%6.2f      %4s%2s%2s\n"

_CONECT_FORMAT_STRING = "CONECT%5s%5s%5s%5s%5s\n"


class Select(object):
    """Select everything fo PDB output (for use as a bas class).

    Default selection (everything) during writing - can be used as base class
    to implement selective output. This selects which entities will be written out.
    """

    def __repr__(self):
        return "<Select all>"

    def accept_model(self, model):
        """Overload this to reject models for output."""
        return 1

    def accept_chain(self, chain):
        """Overload this to reject chains for output."""
        return 1

    def accept_residue(self, residue):
        """Overload this to reject residues for output."""
        return 1

    def accept_atom(self, atom):
        """Overload this to reject atoms for output."""
        return 1


class PDBIO(object):
    """Write a Structure object (or a subset of a Structure object) as a string.

    Example:

        >>> p=PDBParser()
        >>> s=p.get_structure("1fat", "1fat.pdb")
        >>> io=PDBIO()
        >>> io.set_structure(s)
        >>> io.save("out.pdb")

    """

    def __init__(self, use_model_flag=0):
        """Creat the PDBIO object.

        @param use_model_flag: if 1, force use of the MODEL record in output.
        @type use_model_flag: int
        """
        self.use_model_flag = use_model_flag

    # private mathods

    def _get_atom_line(self, atom, hetfield, segid, atom_number, resname,
                       resseq, icode, chain_id, charge="  "):
        """Returns an ATOM PDB string (PRIVATE)."""
        if hetfield != " ":
            record_type = "HETATM"
        else:
            record_type = "ATOM  "

        if atom.element:
            element = atom.element.strip().upper()
            if element.capitalize() not in atom_weights:
                raise ValueError("Unrecognised element %r" % atom.element)
            element = element.rjust(2)
        else:
            element = "  "

        name = atom.get_fullname().strip()

        # Pad atom name if:
        #     - smaller than 4 characters
        # AND - is not C, N, O, S, H, F, P, ..., one letter elements
        # AND - first character is NOT numeric (funky hydrogen naming rules)
        if len(name) < 4 and name[:1].isalpha() and len(element.strip()) < 2:
            name = " " + name

        altloc = atom.get_altloc()
        x, y, z = atom.get_coord()
        bfactor = atom.get_bfactor()
        occupancy = atom.get_occupancy()
        try:
            occupancy_str = "%6.2f" % occupancy
        except TypeError:
            if occupancy is None:
                occupancy_str = " " * 6
                import warnings
                from Bio import BiopythonWarning
                warnings.warn("Missing occupancy in atom %s written as blank" %
                              repr(atom.get_full_id()), BiopythonWarning)
            else:
                raise TypeError("Invalid occupancy %r in atom %r"
                                % (occupancy, atom.get_full_id()))

        args = (record_type, atom_number, name, altloc, resname, chain_id,
                resseq, icode, x, y, z, occupancy_str, bfactor, segid,
                element, charge)
        return _ATOM_FORMAT_STRING % args

    # Public methods

    def set_structure(self, pdb_object):
        # MODBY: Alexandre Fassio
        # Get the Structure object.
        parent_struct = pdb_object.get_parent_by_level(level='S')
        # Save a copy of the CONECT records from structure.
        conects = parent_struct.conects

        # Check what the user is providing and build a structure appropriately
        if pdb_object.level == "S":
            structure = pdb_object
        else:
            sb = StructureBuilder()
            # MODBY: Alexandre
            # Keeps the id from the original structure.
            # MODBY: Alexandre
            # Now it pass the PDB file as a parameter
            sb.init_structure(parent_struct.id, parent_struct.pdb_file)
            sb.init_seg(' ')
            # Build parts as necessary
            if pdb_object.level == "M":
                sb.structure.add(pdb_object)
                self.structure = sb.structure
            else:
                sb.init_model(0)
                if pdb_object.level == "C":
                    sb.structure[0].add(pdb_object)
                else:
                    sb.init_chain('A')
                    if pdb_object.level == "R":
                        parent_id = pdb_object.parent.id
                        try:
                            sb.structure[0]['A'].id = parent_id
                        except Exception:
                            pass

                        # MODBY: Alexandre Fassio
                        # The old code (sb.structure[0]['A'].add(pdb_object)) used to try to access a chain whose id is always 'A'.
                        # However, when the informed residue has a different chain id, this access will raise a KeyError exception.
                        # Thus, we should always access the new chain id by using the parent_id variable.
                        sb.structure[0][parent_id].add(pdb_object)
                    else:
                        # Atom
                        sb.init_residue('DUM', ' ', 1, ' ')
                        parent_id = pdb_object.parent.parent.id
                        try:
                            sb.structure[0]['A'].id = parent_id
                        except Exception:
                            pass

                        # MODBY: Alexandre Fassio
                        # The old code (sb.structure[0]['A'].child_list[0].add(pdb_object)) used to try to access a chain whose id
                        # is always 'A'. However, when the informed atom has a different chain id, this access will raise a
                        # KeyError exception. Thus, we should always access the new chain id by using the parent_id variable.
                        sb.structure[0][parent_id].child_list[0].add(pdb_object)

            # Return structure
            structure = sb.structure

        # MODBY: Alexandre Fassio
        # Set the copied CONECTs to the new Structure object.
        structure.conects = conects
        self.structure = structure

    def save(self, file, select=Select(), write_conects=False,
             write_end=True, preserve_atom_numbering=False):
        """
        @param select: selects which entities will be written.
        @type select: object

        Typically select is a subclass of L{Select}, it should
        have the following methods:

         - accept_model(model)
         - accept_chain(chain)
         - accept_residue(residue)
         - accept_atom(atom)

        These methods should return 1 if the entity is to be
        written out, 0 otherwise.

        Typically select is a subclass of L{Select}.
        """
        get_atom_line = self._get_atom_line
        if isinstance(file, basestring):
            fp = open(file, "w")
            close_file = 1
        else:
            # filehandle, I hope :-)
            fp = file
            close_file = 0

        # MODBY: Alexandre Fassio
        # Atom serial numbers to save on the PDB, i.e., not filtered atoms.
        valid_serial_numbers = set()
        serial_number_mapping = {}

        # multiple models?
        if len(self.structure) > 1 or self.use_model_flag:
            model_flag = 1
        else:
            model_flag = 0
        for model in self.structure.get_list():
            if not select.accept_model(model):
                continue
            # necessary for ENDMDL
            # do not write ENDMDL if no residues were written
            # for this model
            model_residues_written = 0
            if not preserve_atom_numbering:
                atom_number = 1
            if model_flag:
                fp.write("MODEL      %s\n" % model.serial_num)
            for chain in model.get_list():
                if not select.accept_chain(chain):
                    continue
                chain_id = chain.get_id()
                # necessary for TER
                # do not write TER if no residues were written
                # for this chain
                chain_residues_written = 0
                for residue in chain.get_unpacked_list():
                    if not select.accept_residue(residue):
                        continue
                    hetfield, resseq, icode = residue.get_id()
                    resname = residue.get_resname()
                    segid = residue.get_segid()
                    for atom in residue.get_unpacked_list():
                        if select.accept_atom(atom):
                            chain_residues_written = 1
                            model_residues_written = 1

                            if preserve_atom_numbering:
                                atom_number = atom.get_serial_number()

                            # MODBY: Alexandre Fassio
                            # Create a mapping between the current serial number and the new serial number.
                            # If preserve_atom_numbering is set to True, this mapping is irrelevant as both
                            # serial numbers will be the same.
                            serial_number_mapping[atom.get_serial_number()] = atom_number

                            s = get_atom_line(atom, hetfield, segid,
                                              atom_number, resname,
                                              resseq, icode, chain_id)

                            # MODBY: Alexandre Fassio
                            # This atom will not be filtered.
                            valid_serial_numbers.add(atom_number)
                            fp.write(s)
                            if not preserve_atom_numbering:
                                atom_number += 1

                if chain_residues_written:
                    # The serial number of the TER record is always one number greater than the serial number of the ATOM/HETATM
                    # preceding the TER.
                    last_serial_num = atom_number
                    if not preserve_atom_numbering:
                        # Now, increment the current serial number in case there is a new chain.
                        atom_number += 1

                    fp.write("TER   %5i      %3s %c%4i%c                                                      \n"
                             % (last_serial_num, resname, chain_id, resseq, icode))

            if model_flag and model_residues_written:
                fp.write("ENDMDL\n")

        # MODBY: Alexandre Fassio
        # Print CONECT records
        if write_conects:
            conects = self.structure.conects

            for serial_number in sorted(conects):
                # Substitutes the old serial number by the new serial number
                new_serial_number = serial_number_mapping.get(serial_number, None)

                if new_serial_number not in valid_serial_numbers:
                    continue

                # It may return a list of lists as some atoms may have more than one CONECT line.
                bonded_atoms = conects[serial_number]
                max_num_fields = 4
                bonded_atoms_sets = [bonded_atoms[i:i + max_num_fields]
                                     for i in range(0, len(bonded_atoms),
                                                    max_num_fields)]

                for bonded_atoms in bonded_atoms_sets:
                    valid_bonded_atoms = []
                    for tmp_serial_number in bonded_atoms:
                        # Substitutes the old serial number by the new serial number
                        tmp_serial_number = serial_number_mapping.get(tmp_serial_number, None)

                        if tmp_serial_number in valid_serial_numbers:
                            valid_bonded_atoms.append(tmp_serial_number)

                    if len(valid_bonded_atoms) == 0:
                        continue

                    valid_bonded_atoms = [str(x) for x in valid_bonded_atoms]
                    missing_values = max_num_fields - len(valid_bonded_atoms)
                    valid_bonded_atoms += [''] * missing_values

                    record = _CONECT_FORMAT_STRING % (str(new_serial_number),
                                                      *valid_bonded_atoms)

                    fp.write(record)
        if write_end:
            fp.write('END\n')
        if close_file:
            fp.close()


if __name__ == "__main__":

    from luna.MyBio.PDB.PDBParser import PDBParser

    import sys

    p = PDBParser(PERMISSIVE=True)

    s = p.get_structure("test", sys.argv[1])

    io = PDBIO()
    io.set_structure(s)
    io.save("out1.pdb")

    with open("out2.pdb", "w") as fp:
        s1 = p.get_structure("test1", sys.argv[1])
        s2 = p.get_structure("test2", sys.argv[2])
        io = PDBIO(1)
        io.set_structure(s1)
        io.save(fp)
        io.set_structure(s2)
        io.save(fp, write_end=1)



from rdkit.Chem import SanitizeFlags, SanitizeMol
from openbabel.openbabel import OBSmartsPattern

from luna.wrappers.base import BondType, AtomWrapper, MolWrapper
from luna.mol.charge_model import OpenEyeModel

import logging

logger = logging.getLogger()


class MolValidator:
    """Validate and fix molecules with the errors most commonly found when parsing PDB files with Open Babel.

    Parameters
    ----------
    charge_model : :class:`~luna.mol.charge_model.ChargeModel`
        A charge model object. By default, the implementation of OpenEye charge model is used.
    fix_nitro : bool
        If True, fix nitro groups whose nitrogens are perceived as having a valence of 5 and
        that are double-bonded to the oxygens.
        SMILES representation of the error: '[$([NX3v5]([!#8])(=O)=O)]'.
    fix_amidine_and_guanidine : bool
        If True, fix amidine and guanidine-like groups.
        When the molecule is ionized, it may happen that the charge is incorrectly assigned to the
        central carbon, which ends up with a +1 charge and the nitrogen double-bonded
        to it ends up with a +0 charge.
    fix_valence : bool
        If True, fix the valence of atoms.
        Currently, we only detect and fix errors of quaternary ammonium nitrogens.
        These charged atoms are sometimes perceived as having a valence equal to 5 and no charge,
        i.e., Open Babel considers those nitrogens as hypervalent.
    fix_charges : bool
        If True, fix charges on basis of the charge model ``charge_model``.
    """

    def __init__(self, charge_model=OpenEyeModel(), fix_nitro=True, fix_amidine_and_guanidine=True,
                 fix_valence=True, fix_charges=True):
        self.charge_model = charge_model
        self.fix_nitro = fix_nitro
        self.fix_amidine_and_guanidine = fix_amidine_and_guanidine
        self.fix_valence = fix_valence
        self.fix_charges = fix_charges

    def validate_mol(self, mol_obj):
        """Validate molecule ``mol_obj``.

        It will first try to fix the provided molecule according to the initialization flags.
        If any errors are detected and if it fails to fix them, then this function may return False,
        i.e., the molecule is invalid.

        Parameters
        ----------
        mol_obj : :class:`~luna.wrappers.base.MolWrapper`, :class:`rdkit.Chem.rdchem.Mol`, or :class:`openbabel.pybel.Molecule`
            The molecule.

        Returns
        -------
        is_valid : bool
            If the molecule is valid or not.
        """

        if not isinstance(mol_obj, MolWrapper):
            mol_obj = MolWrapper(mol_obj)

        # Validations to be made...
        #       TODO: check aromaticity.
        #       TODO: return list of errors

        if self.fix_nitro:
            self._fix_nitro_substructure_and_charge(mol_obj)

        if self.fix_amidine_and_guanidine:
            self._fix_amidine_and_guanidine_charges(mol_obj)

        # Check if the molecule has errors...
        is_valid = True
        for atm_obj in mol_obj.get_atoms():
            if not self._is_valence_valid(atm_obj):
                is_valid = False

            if not self._is_charge_valid(atm_obj):
                is_valid = False

        if not is_valid:
            logger.debug("Invalid molecule: check the logs for more information.")

        return is_valid

    def _fix_nitro_substructure_and_charge(self, mol_obj):
        ob_smart = OBSmartsPattern()
        # Invalid nitro pattern.
        ob_smart.Init("[$([NX3v5]([!#8])(=O)=O)]")
        if ob_smart.Match(mol_obj.unwrap()):
            logger.debug("One or more invalid nitro substructures ('*-N(=O)=O') were found. "
                         "It will try to substitute them to '*-[N+]([O-])=O'.")

            # Iterate over each Nitro group in the molecule.
            for ids in ob_smart.GetUMapList():
                # Get the N atom.
                atm_obj = [AtomWrapper(mol_obj.GetAtom(i)) for i in ids][0]
                for bond in atm_obj.get_bonds():
                    partner = bond.get_partner_atom(atm_obj)

                    if partner.get_symbol() == "O" and bond.get_bond_type() == BondType.DOUBLE:
                        # Change double bond to single bond.
                        bond.set_bond_type(BondType.SINGLE)
                        # Attributes a +1 charge to the N.
                        atm_obj.set_charge(1)
                        # Attributes a -1 charge to the O.
                        partner.set_charge(-1)

                        # It needs to update only one of the oxygen bonds.
                        break

            ob_smart = OBSmartsPattern()
            # Valid nitro pattern.
            ob_smart.Init("[$([NX3v4+](=O)[O-])][!#8]")
            if ob_smart.Match(mol_obj.unwrap()):
                logger.debug("Invalid nitro substructures ('*-N(=O)=O') successfully substituted to '*-[N+]([O-])=O'.")

    def _fix_amidine_and_guanidine_charges(self, mol_obj):
        # These errors occur with guanidine-like substructures when the molecule is ionized. It happens that the charge is
        # incorrectly assigned to the central carbon, so the guanidine-like C ends up with a +1 charge and the N with a double bond
        # to the central C ends up with a +0 charge. To fix it, we assign the correct charges to the N (+1) and C (0).

        ob_smart = OBSmartsPattern()
        # Invalid amidine and guanidine pattern.
        ob_smart.Init("[$([NH1X2v3+0](=[CH0X3+1](N)))]")
        if ob_smart.Match(mol_obj.unwrap()):
            logger.debug("One or more amidine/guanidine substructures with no charge were found. "
                         "It will try to attribute a +1 charge to the N bound to the central carbon with a double bond.")

            # Iterate over each Amidine/Guanidine group in the molecule.
            for ids in ob_smart.GetUMapList():
                # Get the N atom.
                atm_obj = [AtomWrapper(mol_obj.GetAtom(i)) for i in ids][0]

                for bond in atm_obj.get_bonds():
                    partner = bond.get_partner_atom(atm_obj)

                    if partner.get_symbol() == "C":
                        # Attributes a +1 charge to the N and corrects its degree to three, i.e., the N will become '=NH2+'.
                        atm_obj.set_charge(1)

                        # Set the number of implicit Hydrogens to 1 (=NH2+).
                        #
                        # Explanation: the above initialized Smarts pattern checks for nitrogens containing 1 explicit
                        #              hydrogen ([NH1X2v3+0]). Thus, to correctly update its valence to 4 (=NH2+),
                        #              we need to add a new implicit hydrogen.
                        atm_obj.unwrap().SetImplicitHCount(1)

                        # Remove any charges in the C.
                        partner.set_charge(0)

            ob_smart = OBSmartsPattern()
            # Valid amidine and guanidine pattern.
            ob_smart.Init("[$([NH2X3v4+1](=[CH0X3+0](N)))]")
            if ob_smart.Match(mol_obj.unwrap()):
                logger.debug("Invalid amidine/guanidine substructures were correctly charged.")

    def _is_valence_valid(self, atm_obj):
        if not isinstance(atm_obj, AtomWrapper):
            atm_obj = AtomWrapper(atm_obj)

        # Currently, atoms other than N are not evaluated as no valence error has been identified to them.
        if atm_obj.get_atomic_num() == 7:
            # It corrects quaternary ammonium Nitrogen errors.
            #
            #   While reading from PDBs containing quaternary ammonium N, it may happen to the N to be perceived as
            #       having a valence equal to 5 (v5, hypervalent). It means Open Babel has added an invalid implicit hydrogen.
            #
            if atm_obj.get_valence() == 5 and atm_obj.get_charge() == 0:
                logger.debug("Atom # %d has incorrect valence and charge." % atm_obj.get_id())

                if self.fix_valence:
                    logger.debug("'Fix valence' option is set on. It will update the valence of atom # %d "
                                 "from %d to 4 and correct its charge." % (atm_obj.get_id(), atm_obj.get_valence()))

                    # Set the number of implicit hydrogens to 0 and adds a +1 charge to the Nitrogen.
                    # It is necessary because Open Babel tends to add 1 implicit hydrogen to ammonium nitrogens
                    # what makes them become hypervalent (v5) and with a neutral charge.
                    atm_obj.unwrap().SetImplicitHCount(0)
                    atm_obj.set_charge(1)
                    return True
                return False
        return True

    def _is_charge_valid(self, atm_obj):
        if not isinstance(atm_obj, AtomWrapper):
            atm_obj = AtomWrapper(atm_obj)

        expected_charge = self._get_expected_charge(atm_obj)

        if expected_charge is not None and expected_charge != atm_obj.get_charge():
            logger.debug("Atom # %d has incorrect charges defined." % atm_obj.get_id())

            if self.fix_charges:
                logger.debug("'Fix charges' option is set on. It will update the charge of atom # %d from %d to %d."
                             % (atm_obj.get_id(), atm_obj.get_charge(), expected_charge))
                atm_obj.set_charge(expected_charge)
                return True
            else:
                return False
        return True

    def _get_expected_charge(self, atm_obj):
        if not isinstance(atm_obj, AtomWrapper):
            atm_obj = AtomWrapper(atm_obj)

        return self.charge_model.get_charge(atm_obj)


class RDKitValidator:
    """Check if RDKit molecular objects are valid or not.

    Parameters
    ----------
    sanitize_opts : :class:`rdkit.Chem.rdchem.SanitizeFlags`
        Sanitization operations to be carried out.
    """

    def __init__(self, sanitize_opts=SanitizeFlags.SANITIZE_ALL):
        self.sanitize_opts = sanitize_opts

    def is_valid(self, rdk_mol):
        """Try to sanitize the molecule ``rdk_mol``. If it succeeds, returns True. Otherwise, returns False.

        Parameters
        ----------
        rdk_mol : :class:`rdkit.Chem.rdchem.Mol`

        Returns
        -------
        is_valid : bool
            If the molecule is valid or not.
        """
        try:
            SanitizeMol(rdk_mol, sanitizeOps=self.sanitize_opts)
            return True
        except Exception as e:
            logger.exception(e)
            return False
